import torch
import logging
import cv2
import matplotlib.pyplot as plt
import numpy as np

class Utils():

    @staticmethod
    def get_device(logger):

        if torch.cuda.is_available():
            device = torch.device("cuda")
            logger.info("Device available in CUDA. Using GPU.")
        else:
            device = torch.device("cpu")
            logger.info("Could not find device in CUDA. Using CPU.")

        return device
    
    @staticmethod
    def params_to_range(range_params):
        start = range_params["min"]
        stop = range_params["max"]
        step = range_params["step"]
        
        if isinstance(start, int) and isinstance(stop, int) and isinstance(step, int):
            return range(start, stop + 1, step)
        
        elif isinstance(start, (int, float)) and isinstance(stop, (int, float)) and isinstance(step, (int, float)):
            return np.arange(start, stop + step, step)
    
    @staticmethod
    def overlay_mask_on_image(image_path, mask, alpha=0.5):
        # Cargar imagen original
        img = cv2.imread(image_path)
        h, w, _ = img.shape  # Obtener dimensiones

        # Asegurar que la máscara tenga el mismo tamaño que la imagen
        if mask.shape[:2] != (h, w):
            print(f"Mask shape is {h}x{w} while imgs shape is {img.shape[0]}x{img.shape[1]}")
            mask = cv2.resize(mask, (w, h), interpolation=cv2.INTER_NEAREST)

        # Obtener clases únicas en la máscara
        unique_classes = np.unique(mask)
        unique_classes = unique_classes[unique_classes > 0]  # Excluir fondo (0)

        # Asignar colores aleatorios a cada clase
        np.random.seed(42)  # Para mantener consistencia en los colores
        colors = {class_id: np.random.randint(0, 255, size=(3,), dtype=np.uint8) for class_id in unique_classes}

        # Crear una máscara en color
        mask_colored = np.zeros((*mask.shape, 3), dtype=np.uint8)

        for class_id, color in colors.items():
            mask_colored[mask == class_id] = color

        # Mezclar imagen original con la máscara coloreada
        overlay = cv2.addWeighted(img, 1 - alpha, mask_colored, alpha, 0)

        # Mostrar la imagen resultante
        plt.figure(figsize=(10, 5))
        plt.imshow(overlay)
        plt.axis('off')
        plt.title("Máscara sobrepuesta con colores por clase")
        plt.show()

        return overlay



import numpy as np
import pandas as pd
import os
import shutil
import json
from tqdm import tqdm

from skmultilearn.model_selection import IterativeStratification
from sklearn.model_selection import KFold, train_test_split

from sklearn.utils import shuffle

from segdan.stratification.stratification_fn import calculate_object_number, calculate_pixel_ratio, calculate_pixel_distribution
from segdan.utils.constants import StratificationStrategy

class TrainingDataset():

    def __init__(self, img_path, multilabel_label_path, output_path, label_format, original_label_path):
        self.img_path = img_path
        self.label_path = multilabel_label_path
        self.output_path = output_path
        self.label_format = label_format
        self.original_label_path = original_label_path

        self.image_files = sorted([os.path.join(img_path, f) for f in os.listdir(img_path) if f.lower().endswith(('.jpg', '.jpeg', '.png', '.bmp', '.tiff'))])
        
        self.mask_paths_multilabel = sorted([os.path.join(multilabel_label_path, f) for f in os.listdir(multilabel_label_path) if f.endswith('.png')])
    
    def save_mask_format(self, row, img_dir, label_dir):
        shutil.copy(row["img_path"], os.path.join(img_dir, os.path.basename(row["img_path"])))
        shutil.copy(row["mask_path"], os.path.join(label_dir, os.path.basename(row["mask_path"])))
    
    def save_txt_format(self, row, img_dir, label_dir):
        shutil.copy(row["img_path"], os.path.join(img_dir, os.path.basename(row["img_path"])))
        txt_label_path = row["mask_path"].replace(".png", ".txt").replace(".jpg", ".txt")
        txt_label_path = os.path.join(self.original_label_path, os.path.basename(txt_label_path))
        if os.path.exists(txt_label_path):
            shutil.copy(txt_label_path, os.path.join(label_dir, os.path.basename(txt_label_path)))

    def save_json_format(self, row, img_dir, coco_data, selected_images, selected_annotations, image_ids):
        img_name = os.path.basename(row["img_path"])
        for img in coco_data["images"]:
            if img["file_name"] == img_name:
                selected_images.append(img)
                current_img_id = img["id"]
                image_ids.add(current_img_id)
                break
        for ann in coco_data["annotations"]:
            if ann["image_id"] == current_img_id:
                selected_annotations.append(ann)
        shutil.copy(row["img_path"], os.path.join(img_dir, os.path.basename(row["img_path"])))        

    def save_split_to_directory(self, train_df, val_df, test_df, fold_dir: str = None):

        print("Saving splits...")

        subsets = {"train": train_df, "val": val_df, "test": test_df}

        format_handlers = {
            "mask": self.save_mask_format,
            "txt": self.save_txt_format,
            "json": None  
        }

        output_path = self.output_path

        if fold_dir:
            output_path = fold_dir

        for subset, df in subsets.items():
            if df is None:
                continue
            
            img_dir = os.path.join(output_path, subset, "images")
            label_dir = os.path.join(output_path, subset, "labels")
            os.makedirs(img_dir, exist_ok=True)
            os.makedirs(label_dir, exist_ok=True)

            if self.label_format in format_handlers:
                handler = format_handlers[self.label_format]
                if handler:
                    tqdm.pandas(desc=f"Saving {subset} images and labels")
                    df.progress_apply(lambda row: handler(row, img_dir, label_dir), axis=1)
                    continue  
                    
            with open(self.original_label_path) as f:
                coco_data = json.load(f)

            selected_images = []
            selected_annotations = []
            image_ids = set()
            
            tqdm.pandas(desc=f"Saving {subset} JSON annotations")
            df.progress_apply(lambda row: self.save_json_format(row, img_dir, label_dir, coco_data, selected_images, selected_annotations, image_ids), axis=1)
        
            coco_subset = {
                "images": selected_images,
                "annotations": selected_annotations,
                "categories": coco_data["categories"]
            }
            with open(os.path.join(label_dir, f"{subset}.json"), "w") as f:
                json.dump(coco_subset, f, separators=(',', ':'))

        print(f"Splits successfully saved in {output_path}")
        
    def split(self, general_data: dict, split_data: dict):

        stratification_strategy = split_data["stratification_type"]
        stratify = split_data["stratification"]
        hold_out = split_data["split_method"]
        
        random_state = split_data["stratification_random_seed"]
        background = general_data["background"]
        n_folds = split_data["cross_val"]["num_folds"]

        if stratify and stratification_strategy.lower() not in ["pixels", "objects", "pixel_to_object_ratio"]:
            raise ValueError("Invalid value for stratification_strategy. Must be 'pixels', 'objects' or 'pixel_to_object_ratio'.")
        
        if hold_out:
            train_fraction = split_data["hold_out"]["train"]
            val_fraction = split_data["hold_out"]["valid"]
            test_fraction = split_data["hold_out"]["test"]
            if not np.isclose(train_fraction + val_fraction + test_fraction, 1.0):
                raise ValueError("The sum of train_fraction, val_fraction, and test_fraction must equal 1.0")

        if self.mask_paths_multilabel != None and len(self.image_files) != len(self.mask_paths_multilabel):
            raise ValueError("The number of images and masks must be the same.")
        
        if hold_out: 
            num_classes = None
            if stratify:
                train_images, train_masks, val_images, val_masks, test_images, test_masks, num_classes = self.stratify_split(train_fraction, val_fraction, test_fraction, stratification_strategy, background, random_state)
            else:
                train_images, train_masks, val_images, val_masks, test_images, test_masks = self.random_split(train_fraction, val_fraction, test_fraction, random_state)
        
            train_df = pd.DataFrame({"img_path": train_images, "mask_path": train_masks})
            val_df = pd.DataFrame({"img_path": val_images, "mask_path": val_masks})
            test_df = pd.DataFrame({"img_path": test_images, "mask_path": test_masks})

            self.save_split_to_directory(train_df, val_df, test_df)

            return num_classes
        
        return self.kfold_cross_validation(n_folds, stratify, stratification_strategy, background, random_state)

        
    def stratify_split(self, train_fraction, val_fraction, test_fraction, stratification_strategy, background, random_state):
        """
        Stratify-shuffle-split a semantic segmentation dataset into
        train/val/test sets based on pixel-wise class distributions and save
        the results in specified directories.

        Args:
            image_paths: List of file paths to the input images.
            mask_paths: List of file paths to the corresponding segmentation masks.
            train_fraction: Fraction of data to reserve for the training dataset.
            val_fraction: Fraction of data to reserve for the validation dataset.
            test_fraction: Fraction of data to reserve for the test dataset.
            output_dir: Parent directory where the train, val, and test directories will be created. If None, the directories will not be created. Defaults to None.

        Returns:
            Tuple containing three DataFrames: train, val, and test subsets.
            Each DataFrame has two columns: 'img_path' and 'mask_path'.
        """
        image_paths, mask_paths = shuffle(self.image_files, self.mask_paths_multilabel, random_state=random_state)

        print("Starting classes stratification...")
        if stratification_strategy.lower() == StratificationStrategy.PIXELS.value:
            result = calculate_pixel_distribution(mask_paths, background)
        elif stratification_strategy.lower() == StratificationStrategy.OBJECTS.value:
            result = calculate_object_number(mask_paths, background)
        else:
            result = calculate_pixel_ratio(mask_paths, background)
        
        distributions = result["distributions"]
        num_classes = result["num_classes"]

        print(f"Stratification done. {num_classes} classes detected.")
        
        stratifier = IterativeStratification(
            n_splits=2,
            order=1,
            sample_distribution_per_fold=[train_fraction, 1.0 - train_fraction]
        )

        everything_else_indexes, train_indexes = next(stratifier.split(X=np.zeros(len(image_paths)), y=distributions))

        train_images = [image_paths[i] for i in train_indexes]
        train_masks = [mask_paths[i] for i in train_indexes]

        everything_else_images = [image_paths[i] for i in everything_else_indexes]
        everything_else_masks = [mask_paths[i] for i in everything_else_indexes]

        val_proportion = val_fraction / (val_fraction + test_fraction)  

        stratifier = IterativeStratification(
            n_splits=2,
            order=1,
            sample_distribution_per_fold=[val_proportion, 1.0 - val_proportion]
        )

        test_indexes, val_indexes  = next(stratifier.split(X=np.zeros(len(everything_else_images)), y=distributions[everything_else_indexes]))

        val_images = [everything_else_images[i] for i in val_indexes]
        val_masks = [everything_else_masks[i] for i in val_indexes]
        test_images = [everything_else_images[i] for i in test_indexes]
        test_masks = [everything_else_masks[i] for i in test_indexes]

        return train_images, train_masks, val_images, val_masks, test_images, test_masks, num_classes
    
    def random_split(self, train_fraction, val_fraction, test_fraction, random_state):
        
        train_val_images, test_images, train_val_masks, test_masks = train_test_split(
            self.image_files, self.mask_paths_multilabel, test_size=test_fraction, random_state=random_state
        )

        val_fraction_adjusted = val_fraction / (train_fraction + val_fraction)  
        train_images, val_images, train_masks, val_masks = train_test_split(
            train_val_images, train_val_masks, test_size=val_fraction_adjusted, random_state=random_state
        )

        return train_images, train_masks, val_images, val_masks, test_images, test_masks
    
    
    def kfold_cross_validation(
        self,
        n_splits: int,
        stratify: bool = False, 
        stratification_strategy: str = "pixels",
        background: int = None,
        random_state: int = 42
        ):
        """
        Automates the partitioning process into a fixed test set + K-Fold for train/val.

        Args:
            output_dir: Directory where the partitions will be saved. If None, they are not saved.
            test_fraction: Fraction of the dataset allocated to the test set.
            n_splits: Number of folds for K-Fold split in train/val.
            random_state: Seed for randomization.

        Returns:
            - test_df: DataFrame containing the test set.
            - folds: List of tuples [(train_df, val_df)] for each fold.
        """
        if len(self.image_files) != len(self.mask_paths_multilabel):
            raise ValueError("The number of images and masks must be the same.")

        image_paths, mask_paths = shuffle(self.image_files, self.mask_paths_multilabel, random_state=random_state)

        used_val_indices = set()
        folds = []
        num_classes = None

        if not stratify:
            print("Saving splits...")

            kfold = KFold(n_splits=n_splits, shuffle=True, random_state=random_state)

            for fold_idx, (train_idx, val_idx) in tqdm(enumerate(kfold.split(X=np.zeros(len(image_paths)), y=np.zeros(len(image_paths)))), 
                                                       total=n_splits, 
                                                       desc="Processing folds"):

                if any(idx in used_val_indices for idx in val_idx):
                    raise ValueError(f"Duplicate indices found in fold {fold_idx + 1} for the validation set.")
                
                used_val_indices.update(val_idx)
                
                train_images = [image_paths[i] for i in train_idx]
                train_masks = [mask_paths[i] for i in train_idx]
                val_images = [image_paths[i] for i in val_idx]
                val_masks = [mask_paths[i] for i in val_idx]

                train_df = pd.DataFrame({"img_path": train_images, "mask_path": train_masks})
                val_df = pd.DataFrame({"img_path": val_images, "mask_path": val_masks})
                folds.append((train_df, val_df))

                fold_dir = os.path.join(self.output_path, f"fold_{fold_idx + 1}")
                self.save_split_to_directory(train_df, val_df, None, fold_dir)
                
            return num_classes

        print("Starting stratification...")

        if stratification_strategy.lower() == StratificationStrategy.PIXELS.value:
            result = calculate_pixel_distribution(mask_paths, background)
        elif stratification_strategy.lower() == StratificationStrategy.OBJECTS.value:
            result = calculate_object_number(mask_paths, background)
        else:
            result = calculate_pixel_ratio(mask_paths, background)

        distributions = result["distributions"]
        num_classes = result["num_classes"]

        print(f"Stratification done. {num_classes} classes detected.")


        stratifier = IterativeStratification(
            n_splits=n_splits,
            order=1,
            sample_distribution_per_fold=[1.0 / n_splits] * n_splits
        )

        for fold_idx, (train_idx, val_idx) in tqdm(enumerate(stratifier.split(X=np.zeros(len(image_paths)), 
                                                                      y=distributions)), 
                                           total=n_splits, 
                                           desc="Processing folds"):

            if any(idx in used_val_indices for idx in val_idx):
                raise ValueError(f"Duplicate indices found in fold {fold_idx + 1} for the validation set.")
            
            used_val_indices.update(val_idx)
            
            train_images = [image_paths[i] for i in train_idx]
            train_masks = [mask_paths[i] for i in train_idx]
            val_images = [image_paths[i] for i in val_idx]
            val_masks = [mask_paths[i] for i in val_idx]

            train_df = pd.DataFrame({"img_path": train_images, "mask_path": train_masks})
            val_df = pd.DataFrame({"img_path": val_images, "mask_path": val_masks})
            folds.append((train_df, val_df))

            fold_dir = os.path.join(self.output_path, f"fold_{fold_idx + 1}")
            self.save_split_to_directory(train_df, val_df, None, fold_dir)

        return num_classes



import albumentations as A

def get_training_augmentation(resize_height = 512, resize_width=512):
    train_transform = [
        A.Resize(height=resize_height, width=resize_width),
        A.HorizontalFlip(p=0.5),
        A.ShiftScaleRotate(
            scale_limit=0.5, rotate_limit=0, shift_limit=0.1, p=1, fill=0, fill_mask=255, border_mode=0
        ),
        A.PadIfNeeded(min_height=None, min_width=None, pad_height_divisor=32, pad_width_divisor=32, fill=0, fill_mask=255),
        A.RandomCrop(height=128, width=128, p=0.5),
        A.GaussNoise(p=0.2),
        A.Perspective(p=0.5),
        A.OneOf(
            [
                A.CLAHE(p=1),
                A.RandomBrightnessContrast(p=1),
                A.RandomGamma(p=1),
            ],
            p=0.9,
        ),
        A.OneOf(
            [
                A.Sharpen(p=1),
                A.Blur(blur_limit=3, p=1),
                A.MotionBlur(blur_limit=3, p=1),
            ],
            p=0.9,
        ),
        A.OneOf(
            [
                A.RandomBrightnessContrast(p=1),
                A.HueSaturationValue(p=1),
            ],
            p=0.9,
        ),
    ]
    return A.Compose(train_transform)


def get_validation_augmentation(resize_height = 512, resize_width=512):
    test_transform = [
       A.Resize(height=resize_height, width=resize_width),
       A.PadIfNeeded(min_height=None, min_width=None, pad_height_divisor=32, pad_width_divisor=32, fill=0, fill_mask=255)
    ]
    return A.Compose(test_transform)


import tkinter as tk
from tkinter import ttk, messagebox

from segdan.forms.modelframe import ModelConfigFrame
from segdan.forms.configfileloaderframe import ConfigFileLoaderFrame
from segdan.forms.datasetsplitframe import DatasetSplitFrame
from segdan.forms.reductionframe import ReductionFrame
from segdan.forms.clusteringconfigframe import ClusteringConfigFrame
from segdan.forms.analysisconfigframe import AnalysisConfigFrame
from segdan.forms.clusteringframe import ClusteringFrame
from segdan.forms.generalconfigframe import GeneralConfigFrame
from segdan.forms.introductionframe import IntroductionFrame
from segdan.forms.reductionconfigframe import ReductionConfigFrame

class Wizard(tk.Tk):
   def __init__(self, *args, **kwargs):
      tk.Tk.__init__(self, *args, **kwargs)

      self.title("SegDAN Wizard")
      self.window_width = 850
      self.window_height = 750

      screen_width = self.winfo_screenwidth()
      screen_height = self.winfo_screenheight()

      x = (screen_width // 2) - (self.window_width // 2)
      y = (screen_height // 2) - (self.window_height // 2)

      self.geometry(f"{self.window_width}x{self.window_height}+{x}+{y}")

      self.resizable(False, False)

      self.config_data = {
         "analyze": tk.BooleanVar(value=False),
         "cluster_images": tk.BooleanVar(value=False),
         "reduce_images": tk.BooleanVar(value=False),
         
      }

      self.container = ttk.Frame(self)
      self.container.grid(row=1, column=0, sticky="nsew")
      self.container.grid_rowconfigure(0, weight=1)
      self.container.grid_columnconfigure(0, weight=1)
      self.final_dict = {}
      
      self.steps = [
         "Introduction", 
         "General configuration", 
         "Analysis", 
         "Clustering", "Clustering configuration", 
         "Reduction", "Reduction configuration", 
         "Dataset split",
         "Model configuration",
         "ConfigFileLoader"
      ]

      self.frames = {}  
      self.current_step = 0

      self.create_breadcrumb()

      steps_classes = [IntroductionFrame, GeneralConfigFrame, AnalysisConfigFrame, ClusteringFrame, ClusteringConfigFrame, ReductionFrame, ReductionConfigFrame, DatasetSplitFrame, ModelConfigFrame, ConfigFileLoaderFrame]

      for Step in steps_classes:
         frame = Step(self.container, self, self.config_data, self.final_dict)
         self.frames[Step.__name__] = frame
         frame.grid(row=0, column=0, sticky="nsew")

      self.show_frame("IntroductionFrame")

      self.protocol("WM_DELETE_WINDOW", self.on_close)

   def create_breadcrumb(self):
      self.breadcrumb_frame = ttk.Frame(self)
      self.breadcrumb_frame.grid(row=0, column=0, padx=10, pady=(5,20))

      self.grid_rowconfigure(0, weight=0)  
      self.grid_rowconfigure(1, weight=1)  
      self.grid_columnconfigure(0, weight=1)  

      self.breadcrumb_labels = []
      for step in self.steps:
         if "configfileloader" in step.lower():
               continue
         
         if "configuration" in step.lower():
            step = step.replace("configuration", "\nconfiguration")
         label = ttk.Label(self.breadcrumb_frame, text=step, padding=(5, 2), justify="center")
         label.pack(side="left", padx=5)
         self.breadcrumb_labels.append(label)

      self.update_breadcrumb()

   def update_breadcrumb(self):
      for i, label in enumerate(self.breadcrumb_labels):
         if i == self.current_step:
            label.config(foreground="blue", font=("Arial", 10, "bold"))
         else:
            label.config(foreground="gray", font=("Arial", 10))

   def on_close(self):
      result = messagebox.askyesno("Close without saving?", "You haven't saved your configuration. Do you really want to exit?")
      if result:  
         self.quit()

   def show_frame(self, cont):
      #print(f"Switching to {cont}, current config_data: {self.config_data}")

      for i, step_name in enumerate(self.frames.keys()):
         if step_name == cont:
            self.current_step = i
            break

      frame = self.frames[cont]
      frame.tkraise()
      self.update_breadcrumb()


import tkinter as tk
from tkinter import ttk, filedialog, messagebox
from tktooltip import ToolTip

import os
import yaml
import json

class ConfigFileLoaderFrame(ttk.Frame):

    def __init__(self, parent, controller, config_data, final_dict):
        ttk.Frame.__init__(self, parent)

        self.top = parent
        self.reduction_var = tk.BooleanVar(value=False)
        self.config_data = config_data
        self.controller = controller
        self.config_path = tk.StringVar(value="")
        self.final_dict = final_dict

        label_title = ttk.Label(self, text="Load configuration settings", font=("Arial", 18, "bold"))
        label_title.grid(row=0, column=0, columnspan=5, pady=(20,10), padx=10)

        self.grid_rowconfigure(0, weight=0)
        self.grid_columnconfigure(0, weight=1)

        self.config_loader_frame = tk.Frame(self, padx=10, pady=10)
        self.config_loader_frame.grid(row=1, column=0, padx=10, pady=10, columnspan=3)

        self.config_loader_frame.grid_rowconfigure(0, weight=0)
        self.config_loader_frame.grid_columnconfigure(0, weight=0)
        self.config_loader_frame.grid_columnconfigure(1, weight=0)

        self.config_file_label = ttk.Label(self.config_loader_frame, text="Configuration file")
        self.config_file_label.grid(row=0, column=0, padx=5, pady=10)

        self.config_file_entry = ttk.Entry(self.config_loader_frame, textvariable=self.config_path, width=50, state="readonly")
        self.config_file_button = tk.Button(self.config_loader_frame, text="Select file 📄", command=lambda: self.select_file())
        ToolTip(self.config_file_button, msg="Select the configuration file (JSON or YAML).")

        self.config_file_load_button = tk.Button(self.config_loader_frame, text="Load file 📄", command=lambda: self.load_file())
        ToolTip(self.config_file_load_button, msg="Load the selected configuration file.")

        self.config_file_entry.grid(row=2, column=0, padx=5, pady=10)
        self.config_file_button.grid(row=2, column=1, padx=5, pady=10)
        self.config_file_load_button.grid(row=2, column=2, padx=5, pady=10)

        self.config_file_load_button.grid_remove()

        button_frame = ttk.Frame(self.config_loader_frame)
        button_frame.grid(row=4, column=0, columnspan=3, pady=10, sticky="e")  

        button_back = ttk.Button(button_frame, text="Back", command=lambda: self.controller.show_frame("IntroductionFrame"))
        button_back.grid(row=0, column=0, padx=50, pady=5, sticky="w")

        button_save = ttk.Button(button_frame, text="Save configuration", command=self.save_configuration)
        button_save.grid(row=0, column=1, pady=5, sticky="w")

        button_frame.grid_columnconfigure(0, weight=0)
        button_frame.grid_columnconfigure(1, weight=0)    

    def select_file(self):
        filetypes = [("YAML Files", "*.yaml;*.yml"),("JSON Files", "*.json")]
        file_path = filedialog.askopenfilename(title="Select File", filetypes=filetypes)
        
        if file_path:
            self.config_path.set(file_path)
            self.toggle_load_button()

    def toggle_load_button(self):
        if self.config_path.get().strip() != "":
            self.config_file_load_button.grid()
        else:
            self.config_file_load_button.grid_remove()

    def load_file(self):
        file_path = self.config_path.get().strip()
        
        if not file_path:
            messagebox.showwarning("No file selected", "Please select a configuration file first.")
            return
        
        file_extension = os.path.splitext(file_path)[1].lower()
        
        try:
            if file_extension == '.json':
                with open(file_path) as json_file:
                    loaded_data = json.load(json_file)
            elif file_extension == '.yaml' or file_extension == '.yml':
                with open(file_path) as yaml_file:
                    loaded_data = yaml.safe_load(yaml_file)
            else:
                messagebox.showerror("Invalid File", "Selected file is neither a JSON nor a YAML file.")
                return
            
            messagebox.showinfo("File Loaded", "Configuration file has been successfully loaded.")

            if 'class_mapping' in loaded_data:
                loaded_data['class_mapping'] = {int(k): v for k, v in loaded_data['class_mapping'].items()}
            
            if 'color_dict' in loaded_data:
                loaded_data['color_dict'] = {int(k): v for k, v in loaded_data['color_dict'].items()}

            self.final_dict.update(loaded_data)

        except Exception as e:
            messagebox.showerror("Error", f"Failed to load the configuration file. Error: {str(e)}")

    def save_configuration(self):
        self.top.quit()




from typing import Optional
import numpy as np
import logging
from transformers.utils import logging as hf_logging

class Converter():

    def __init__(self, input_data: str, output_dir: str):

        self.input_data = input_data
        self.output_dir = output_dir
        
        hf_logging.set_verbosity_error()

        self.logger = logging.getLogger(self.__class__.__name__)
        if not self.logger.hasHandlers(): 
            handler = logging.StreamHandler()
            formatter = logging.Formatter('%(levelname)s - %(message)s', datefmt='%H:%M:%S')
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.logger.setLevel(logging.INFO)

    def _create_empty_mask(self, height: int, width: int, fill_background: Optional[int]):

        if fill_background is None:
          fill_background = 0
                
        return np.full((height, width), fill_background, dtype=np.uint8)
    
    def _scale_polygon(self, polygon, height: int, width: int):
        polygon[:, 0] *= width
        polygon[:, 1] *= height

        return polygon
    
    
    def convert(self):
        raise NotImplementedError("Subclasses must implement this method")



from segdan.converters.converter import Converter
from segdan.utils.imagelabelutils import ImageLabelUtils

import numpy as np
import gc
import os
import cv2
from tqdm import tqdm

class BinaryToMultilabelConverter(Converter):

    def __init__(self, input_data: str, output_dir: str, threshold: int = 255):
        super().__init__(input_data, output_dir)
        self.threshold = threshold

    def convert(self):
        masks = []

        label_files = [f for f in os.listdir(self.input_data) if os.path.isfile(os.path.join(self.input_data, f))]

        os.makedirs(self.output_dir, exist_ok=True)

        for filename in tqdm(label_files, desc="Converting labels from binary format to multilabel..."):
            label_path = os.path.join(self.input_data, filename)
            
            mask = cv2.imread(label_path, cv2.IMREAD_GRAYSCALE)
            
            converted_mask = (mask >= self.threshold).astype(np.uint8)

            ImageLabelUtils.save_multilabel_mask(converted_mask, filename, self.output_dir)
            masks.append(converted_mask)

            del mask, converted_mask
            gc.collect()

        return masks

    





from typing import Optional
import numpy as np
import cv2
import os
import random
from tqdm import tqdm
import matplotlib.pyplot as plt

from segdan.converters.converter import Converter
from segdan.utils.imagelabelutils import ImageLabelUtils

class ColorToMultilabelConverter(Converter):

    num_images_mosaic = 4

    def __init__(self, input_data: str, output_dir: str, color_dict: Optional[dict] = None):
        super().__init__(input_data, output_dir)
        self.color_dict = color_dict

    def generate_color_dict(self):
        label_files = [f for f in os.listdir(self.input_data) if os.path.isfile(os.path.join(self.input_data, f))]
        unique_colors = set()

        for filename in tqdm(label_files, desc="Generating unique colors..."):
            label_path = os.path.join(self.input_data, filename)
            color_mask = cv2.imread(label_path, cv2.IMREAD_UNCHANGED)
            colors = np.unique(color_mask.reshape(-1, color_mask.shape[-1]), axis=0)
            unique_colors.update(map(tuple, colors))

        assigned_colors = {}
        image_dir = self.input_data.replace('labels', 'images')
        
        for color in unique_colors:

            valid_files = []

            for filename in label_files:
                label_path = os.path.join(self.input_data, filename)
                mask = cv2.imread(label_path, cv2.IMREAD_UNCHANGED)
                if np.any(np.all(mask.reshape(-1, mask.shape[-1]) == color, axis=1)):
                    valid_files.append(filename)

            selected_files = random.sample(valid_files, min(self.num_images_mosaic, len(valid_files)))
            fig, axes = plt.subplots(1, self.num_images_mosaic, figsize=(15, 5))
            
            for i, filename in enumerate(selected_files):
                label_path = os.path.join(self.input_data, filename)
                image_path = ImageLabelUtils.label_to_image(label_path, image_dir, '.jpg')

                img = cv2.imread(image_path)
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                mask = cv2.imread(label_path, cv2.IMREAD_UNCHANGED)
                mask_binary = np.all(mask == color, axis=-1).astype(np.uint8)
                mask_colored = np.zeros_like(img)
                mask_colored[mask_binary == 1] = color

                overlay = cv2.addWeighted(img, 0.5, mask_colored, 1, 0)
                axes[i].imshow(overlay)
                axes[i].set_title(filename)
                axes[i].axis("off")
            
            plt.show(block=False)  
            while True:
                try:
                    color_id = int(input(f"Enter class ID for color {color}: "))
                    if color_id not in assigned_colors.values():
                        assigned_colors[color] = color_id
                        break
                    print(f"Invalid input. Class ID already assigned.")
                except ValueError:
                    print("Invalid input. Please enter a valid integer ID.")
            plt.close()  
            
        return assigned_colors

    def convert(self):
        masks = []
        os.makedirs(self.output_dir, exist_ok=True)

        label_files = [f for f in os.listdir(self.input_data) if os.path.isfile(os.path.join(self.input_data, f))]

        if self.color_dict is None:
            print("No color dictionary found. It will be created automatically...")
            self.color_dict = self.generate_color_dict(self.input_data)

        for filename in tqdm(label_files, desc="Converting labels from color format to multilabel..."):
            label_path = os.path.join(self.input_data, filename)
            color_mask = cv2.imread(label_path)
            multilabel_mask = np.zeros(color_mask.shape[:2], dtype=np.uint8)
            pixel_colors = color_mask.reshape(-1, color_mask.shape[-1])

            for idx, pixel_color in enumerate(pixel_colors):
                color_tuple = tuple(pixel_color)
                if color_tuple in self.color_dict:
                    multilabel_mask.reshape(-1)[idx] = self.color_dict[color_tuple]

            multilabel_mask = multilabel_mask.reshape(color_mask.shape[:2])
            masks.append(multilabel_mask)

            ImageLabelUtils.save_multilabel_mask(multilabel_mask, filename, self.output_dir)

        return masks


from imagedatasetanalyzer import KMeansClustering, AgglomerativeClustering, DBSCANClustering, OPTICSClustering
from imagedatasetanalyzer import ImageDataset

import numpy as np

class ClusteringFactory():

    def generate_clustering_model(self, model_name: str, dataset: ImageDataset, embeddings: np.ndarray, random_state: int=123):
        
        models_map = {
            'kmeans': lambda: KMeansClustering(dataset, embeddings, random_state),
            'agglomerative': lambda: AgglomerativeClustering(dataset, embeddings, random_state),
            'dbscan': lambda: DBSCANClustering(dataset, embeddings, random_state),
            'optics': lambda: OPTICSClustering(dataset, embeddings, random_state)
        }

        if model_name not in models_map:
            raise ValueError(f"Clustering model '{model_name}' not supported. ")
        
        return models_map[model_name]()
    



import torch
import warnings

def _handle_zero_division(x, zero_division):
    nans = torch.isnan(x)
    if torch.any(nans) and zero_division == "warn":
        warnings.warn("Zero division in metric calculation!")
    elif zero_division == "ignore":
        return x[~nans]
    
    value = zero_division if zero_division != "warn" else 0
    value = torch.tensor(value, dtype=x.dtype).to(x.device)
    x = torch.where(nans, value, x)
    return x


def custom_metric(tp, fp, fn, tn, metric_fn, reduction="micro", zero_division="warn"):
    if reduction == "micro":
        tp = tp.sum()
        fp = fp.sum()
        fn = fn.sum()
        tn = tn.sum()
        score = metric_fn(tp, fp, fn, tn)

    elif reduction == "micro-imagewise":
        tp = tp.sum(1)
        fp = fp.sum(1)
        fn = fn.sum(1)
        tn = tn.sum(1)
        score = metric_fn(tp, fp, fn, tn)
        score = _handle_zero_division(score, zero_division)
        score = score.mean()
        
    elif reduction == "none" or reduction is None:
        score = metric_fn(tp, fp, fn, tn)
        score = _handle_zero_division(score, zero_division)
        
    return score


import logging
import tkinter as tk
from tkinter import ttk
import sys
import os
import itertools
import copy
import numpy as np


sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../../")))

from imagedatasetanalyzer import ImageDataset

import json
import glob

from src.reduction.reduction import reduce_dataset 
from src.clustering.clustering import cluster_images, get_embeddings
from src.converters.converterfactory import ConverterFactory
from src.analysis.analysis import analyze_data
from src.forms.wizard import Wizard
from src.datasets.traindataset import TrainingDataset
from src.training.partition import dataset_split
from src.utils.imagelabelutils import ImageLabelUtils


def create_logger():
   logger = logging.getLogger("SegmentationPipeline")
   logger.setLevel(logging.INFO)

   if not logger.hasHandlers():
      handler = logging.StreamHandler()
      formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s', datefmt='%H:%M')
      handler.setFormatter(formatter)
      logger.addHandler(handler)
   
   return logger

def general_pipeline(config_dict, logger):
    transformer = ConverterFactory()
    logger = create_logger()

    general_data = config_dict["general_data"]
    output_path = general_data["output_path"]
    verbose = general_data["verbose"]
    class_map = general_data["class_mapping"]
    image_path = general_data["image_path"]
    label_path = general_data["label_path"]
    label_format = general_data["label_format"]
    
    if config_dict["analyze"]:
        analyze_data(general_data, transformer, output_path, class_map, verbose, logger)

    if config_dict["cluster_images"]:
        clustering_data = config_dict["clustering_data"]
        image_dataset = ImageDataset(image_path)
        
        #embeddings = get_embeddings(clustering_data, image_dataset, verbose, logger)

        embeddings = np.load(r"C:\Users\joortif\Desktop\datasets\Grid\embeddings_melanomac3.npy")

        #np.save(os.path.join(output_path, "embeddings_melanomac3.npy"), embeddings) 
        
        clustering_results = cluster_images(clustering_data, image_dataset, embeddings, output_path, verbose, logger)

        if config_dict["reduce_images"]:
           evaluation_metric = clustering_data["clustering_metric"]
           reduction_data = config_dict["reduction_data"]
           reduction_path = reduce_dataset(reduction_data, clustering_results, evaluation_metric, image_dataset, label_path, embeddings, output_path, verbose, logger)
            
           if reduction_data["use_reduced"]:
              image_path = os.path.join(reduction_path, "images")
              label_path = os.path.join(reduction_path, "labels")

    multilabel_transformation_path = os.path.join(output_path, "transformations", "multilabel")

    if not os.path.exists(multilabel_transformation_path):
        multilabel_transformation_path = label_path
    
    split_output_path = os.path.join(output_path, "split")
    
    split_data= config_dict["split_data"]
    
    dataset = TrainingDataset(image_path, multilabel_transformation_path, split_output_path, label_format, label_path)
    
    num_classes = dataset.split(general_data, split_data)

def grid_search(config_dict, 
                logger,
                reduction_methods=["representative", "diverse", "random"], 
                reduction_percentages=[1.0, 0.75, 0.50, 0.25, 0.10], 
                stratification_methods=["nostrat", "pixels", "objects", "pixel_to_object_ratio"]):  
    
    base_output_path = config_dict["general_data"]["output_path"]

    all_combinations = list(itertools.product(reduction_methods, reduction_percentages, stratification_methods))

    filtered_combinations = []
    full_seen_strats = set()

    for reduction_type, reduction_percentage, strat_method in all_combinations:
        if reduction_percentage == 1.0:
            if strat_method not in full_seen_strats:
                filtered_combinations.append((reduction_type, reduction_percentage, strat_method))
                full_seen_strats.add(strat_method)
            continue
        filtered_combinations.append((reduction_type, reduction_percentage, strat_method))

    for reduction_type, reduction_percentage, strat_method in filtered_combinations:
        config_copy = copy.deepcopy(config_dict)
        config_copy["split_data"]["split_method"] = True

        is_full = reduction_percentage == 1.0

        config_copy["general_data"]["output_path"] = os.path.join(
            base_output_path,
            os.path.join("full", strat_method) if is_full else os.path.join(reduction_type, strat_method, str(reduction_percentage))
        )

        os.makedirs(config_copy["general_data"]["output_path"], exist_ok=True)

        if strat_method == "nostrat":
            config_copy["split_data"]["stratification"] = False
        else:
            config_copy["split_data"]["stratification"] = True
            config_copy["split_data"]["stratification_type"] = strat_method

        if is_full:
            config_copy["cluster_images"] = False
            config_copy["reduce_images"] = False
        else:
            config_copy["cluster_images"] = True
            config_copy["reduce_images"] = True
            config_copy["reduction_data"]["reduction_type"] = reduction_type
            config_copy["reduction_data"]["reduction_percentage"] = reduction_percentage

        print(f"Ejecutando combinación → Reducción: {reduction_type}, Porcentaje: {reduction_percentage}, Estratificación: {strat_method}")
        general_pipeline(config_copy, logger=logger)


if __name__ == "__main__":

    app = Wizard()
    app.mainloop()

    config_dict = app.final_dict

    app.destroy()

    if len(config_dict) == 0:
        sys.exit()
    
    logger = create_logger()

    grid_search(config_dict, logger, reduction_methods=["representative"], 
                reduction_percentages=[1.0, 0.70], 
                stratification_methods=["nostrat","objects"])




   


from segdan.utils.constants import LabelFormat
from segdan.converters.converterfactory import ConverterFactory
from segdan.utils.imagelabelutils import ImageLabelUtils

from imagedatasetanalyzer import ImageLabelDataset, ImageDataset
import os 

def convert_to_mask(label_path, image_path, input_format, general_data, output_path, converter_factory):
    transformations_path = os.path.join(output_path, "transformations", LabelFormat.MASK.value)
    os.makedirs(transformations_path, exist_ok=True)

    print(f"Transforming labels from {input_format} to multilabel. Results are saved in {transformations_path}")

    context = {
        "input_data": label_path,
        "img_dir": image_path,
        "output_dir": transformations_path,
        **general_data  
    }

    converter = converter_factory.get_converter(input_format, LabelFormat.MASK.value, context)
    converter.convert()

    return transformations_path

def _analyze_and_save_results(dataset: ImageDataset, output_path: str, verbose: bool):
    analysis_result_path = os.path.join(output_path, "analysis")
    os.makedirs(analysis_result_path, exist_ok=True)

    print("Starting dataset analysis...")
    dataset.analyze(output=analysis_result_path, verbose=verbose)

    print(f"Dataset analysis ended successfully. Results saved in {analysis_result_path}")

def analyze_data(general_data: dict, transformerFactory: ConverterFactory, output_path:str, class_map: dict,  verbose: bool):

    image_path = general_data["image_path"]
    label_path = general_data["label_path"]
    label_format = general_data["label_format"]
    background = general_data.get("background", None)
    binary = general_data.get("binary", False)

    if label_path is None:
        dataset = ImageDataset(image_path)
        _analyze_and_save_results(dataset, output_path, verbose)
        return 

    if label_format == LabelFormat.MASK.value:
        if binary:
            label_format = LabelFormat.BINARY.value
        elif ImageLabelUtils.all_images_are_color(label_path):
            label_format = LabelFormat.COLOR.value
        
    if label_format != LabelFormat.MASK.value:
        label_path = convert_to_mask(label_path, image_path, label_format, general_data, output_path, transformerFactory)

    dataset = ImageLabelDataset(image_path, label_path, background=background, class_map=class_map)
    _analyze_and_save_results(dataset, output_path, verbose)
    return 
        




import os
import pytorch_lightning as pl
import segmentation_models_pytorch as smp
import torch
from torch.optim import lr_scheduler
import time
import numpy as np
import pandas as pd

from segdan.metrics.compute_metrics import compute_metrics
from segdan.models.semanticsegmentationmodel import SemanticSegmentationModel

class SMPModel(pl.LightningModule, SemanticSegmentationModel):
    def __init__(self, in_channels: int , out_classes: int, metrics: np.ndarray, selection_metric: str, epochs:int, t_max: int, output_path:str, ignore_index: int=255, 
                 model_name: str="unet", encoder_name: str="resnet34", **kwargs):
        
        pl.LightningModule.__init__(self)

        SemanticSegmentationModel.__init__(self, out_classes, epochs, metrics, selection_metric, ignore_index, model_name, encoder_name, output_path)
        super().__init__()
        self.model_name = model_name.replace("-", "")
        self.in_channels = in_channels

        self.model = smp.create_model(
            self.model_name,
            encoder_name=self.model_size,
            in_channels=self.in_channels,
            classes=self.out_classes,
            **kwargs,
        )

        self.t_max = t_max
        
        # Preprocessing parameters for image normalization
        params = smp.encoders.get_preprocessing_params(self.model_size)
        self.number_of_classes = out_classes
        self.binary = out_classes == 1
        self.register_buffer("std", torch.tensor(params["std"]).view(1, 3, 1, 1))
        self.register_buffer("mean", torch.tensor(params["mean"]).view(1, 3, 1, 1))

        if self.out_classes > 1:
            self.loss_mode = smp.losses.MULTICLASS_MODE
        else:
            self.loss_mode = smp.losses.BINARY_MODE

        self.loss_fn = smp.losses.DiceLoss(self.loss_mode, from_logits=True, ignore_index=self.ignore_index)

        # Step metrics tracking
        self.training_step_outputs = []
        self.validation_step_outputs = []
        self.test_step_outputs = []

    def forward(self, image):
        # Normalize image
        image = (image - self.mean) / self.std
        mask = self.model(image)
        return mask
    
    def validate_segmentation_batch(self, image, mask, binary=False):

        assert image.ndim == 4, f"Expected image ndim=4, got {image.ndim}" # [batch_size, channels, H, W]
        h, w = image.shape[2:]
        assert h % 32 == 0 and w % 32 == 0, f"Image dimensions must be divisible by 32, got {h}x{w}"


        if binary: 
            if mask.ndim == 3:
                mask = mask.unsqueeze(1)
            assert mask.ndim == 4, f"Expected binary mask ndim=4, got {mask.ndim}"
            assert mask.max() <= 1.0 and mask.min() >= 0.0, "Binary mask values must be in range [0, 1]"
        else:
            assert mask.ndim == 3, f"Expected multiclass mask ndim=3, got {mask.ndim}"
            mask = mask.long()
        
        return image, mask

    def shared_step(self, batch, stage):
        image, mask = batch

        image, mask = self.validate_segmentation_batch(image, mask, self.binary)

        logits_mask = self.forward(image)

        logits_mask = logits_mask.contiguous()

        loss = self.loss_fn(logits_mask, mask)

        if self.binary:
            prob_mask = logits_mask.sigmoid()
            pred_mask = (prob_mask > 0.5).float()
        
        else:
            prob_mask = logits_mask.softmax(dim=1)
            pred_mask = prob_mask.argmax(dim=1)

        if self.binary:
            metric_args = {"mode": "binary"}
        else:
            metric_args = {"mode": "multiclass", "num_classes": self.number_of_classes}

        if self.ignore_index is not None:
            metric_args["ignore_index"] = self.ignore_index

        tp, fp, fn, tn = smp.metrics.get_stats(pred_mask.long(), mask.long(), **metric_args)
                
        return {
            "loss": loss,
            "tp": tp,
            "fp": fp,
            "fn": fn,
            "tn": tn,
        }

    def shared_epoch_end(self, outputs, stage):
        results = compute_metrics(outputs, self.metrics, stage)

        self.log_dict(results, prog_bar=True)

    def training_step(self, batch):
        train_loss_info = self.shared_step(batch, "train")
        self.training_step_outputs.append(train_loss_info)
        return train_loss_info

    def on_train_epoch_end(self):
        self.shared_epoch_end(self.training_step_outputs, "train")
        self.training_step_outputs.clear()

    def validation_step(self, batch):
        valid_loss_info = self.shared_step(batch, "valid")
        self.validation_step_outputs.append(valid_loss_info)
        return valid_loss_info

    def on_validation_epoch_end(self):
        self.shared_epoch_end(self.validation_step_outputs, "valid")
        self.validation_step_outputs.clear()

    def test_step(self, batch):
        test_loss_info = self.shared_step(batch, "test")
        self.test_step_outputs.append(test_loss_info)
        return test_loss_info

    def on_test_epoch_end(self):
        self.shared_epoch_end(self.test_step_outputs, "test")
        self.test_step_outputs.clear()

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=2e-4)
        scheduler = lr_scheduler.CosineAnnealingLR(optimizer, T_max=self.t_max, eta_min=1e-5)
        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": scheduler,
                "interval": "step",
                "frequency": 1,
            },
        }

    def save_metrics(self,trainer, dataloader, experiment_name, filename="metrics.csv", training_time=None):
        metrics = trainer.evaluate(eval_dataset=dataloader)
        
        if not metrics:
            print("No metrics to save.")
            return metrics

        df = pd.DataFrame(metrics)
        df.insert(0, experiment_name)

        if training_time is not None:
            df["Training Time (min)"] = round(training_time / 60.0, 2)

        if os.path.exists(filename):
            df_existing = pd.read_csv(filename, sep=';')
            df_combined = pd.concat([df_existing, df], ignore_index=True)
        else:
            df_combined = df

        df_combined.to_csv(filename, sep=';', index=False)
        print(f"Metrics saved in file {filename}")

        evaluation_metric = metrics[0].get(f"test_{self.selection_metric}")
        return evaluation_metric

    def train(self, train_loader, valid_loader, test_loader):
        trainer = pl.Trainer(max_epochs=self.epochs, log_every_n_steps=1)
        start_time = time.time()
        trainer.fit(self, train_dataloaders=train_loader, val_dataloaders=valid_loader)
        end_time = time.time()
        total_time = end_time - start_time
        print(f"Total training time: {total_time / 60:.2f} minutes")

        if valid_loader is not None:
            valid_metrics = trainer.validate(self, dataloaders=valid_loader, verbose=False)
            print(valid_metrics)

        evaluation_metric = self.save_metrics(trainer, test_loader, f"{self.model_name} - {self.model_size}", f"metrics_{self.model_name}.csv", training_time=total_time)
        model_output_path = self.save_model(self.output_path)

        return evaluation_metric, model_output_path


from typing import List, Optional
import numpy as np
import pandas as pd
import os
import shutil
from tqdm import tqdm
from skmultilearn.model_selection import IterativeStratification
from sklearn.model_selection import train_test_split

from PIL import Image

from sklearn.utils import shuffle
import matplotlib.pyplot as plt

from imagedatasetanalyzer import ImageLabelDataset

import cv2

if __name__ == "__main__":

    
    def calculate_pixel_distribution(mask_paths, background: Optional[int] = None):
        all_distributions = []
        unique_classes = set()

        for mask_path in tqdm(mask_paths, desc="Calculating pixel distribution for each class"):
            with Image.open(mask_path) as mask_image:
                mask = np.array(mask_image)
                unique, counts = np.unique(mask, return_counts=True)

                if background is not None:
                    mask_exclude_idx = np.where(unique == background)[0]
                    if len(mask_exclude_idx) > 0: 
                        unique = np.delete(unique, mask_exclude_idx)
                        counts = np.delete(counts, mask_exclude_idx)

                unique_classes.update(unique)

                distribution = {class_id: count for class_id, count in zip(unique, counts)}
                all_distributions.append(distribution)
                
        num_classes = len(unique_classes)
        sorted_classes = sorted(unique_classes)

        matrix_distributions = np.zeros((len(mask_paths), num_classes))

        for i, distribution in enumerate(all_distributions):
            for class_id, count in distribution.items():
                class_index = sorted_classes.index(class_id)
                matrix_distributions[i, class_index] = count

        row_sums = matrix_distributions.sum(axis=1, keepdims=True)  
        matrix_distributions = np.divide(matrix_distributions, row_sums, where=row_sums != 0)  
        
        return {
            "distributions": matrix_distributions,
            "num_classes": num_classes
        }

    def calculate_object_number( masks, background: Optional[int] = None):
        all_objects_per_class = []
        unique_classes = set()

        for mask_path in tqdm(masks, desc="Reading number of classes from images"):
            with Image.open(mask_path) as mask_image:
                mask = np.array(mask_image)
                unique = np.unique(mask)

                if background is not None:
                    unique = unique[unique != background]

                unique_classes.update(unique)

        sorted_classes = sorted(unique_classes)  
        num_classes = len(sorted_classes)
        class_mapping = {class_id: idx for idx, class_id in enumerate(sorted_classes)}

        for mask_path in tqdm(masks, desc="Calculating number of objects for each class"):
            with Image.open(mask_path) as mask_image:
                mask = np.array(mask_image)

                objects_per_class = np.zeros(num_classes) 

                for class_id in np.unique(mask):
                    if background is not None and class_id == background:
                        continue

                    if class_id in class_mapping:
                        class_mask = np.where(mask == class_id, 255, 0).astype(np.uint8)
                        contours, _ = cv2.findContours(class_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
                        objects_per_class[class_mapping[class_id]] = len(contours)

                all_objects_per_class.append(objects_per_class)

        return {
            "distributions": np.array(all_objects_per_class),
            "num_classes": num_classes,
            "class_mapping": class_mapping
        }
    
    def calculate_pixel_ratio(masks, background: Optional[int] = None):
        object_info = calculate_object_number(masks, background)
        objects_per_class = object_info["distributions"]  
        num_classes = object_info["num_classes"]
        class_mapping = object_info["class_mapping"]

        all_pixel_ratios = []

        for i, mask_path in tqdm(enumerate(masks), total=len(masks), desc="Calculating pixel-to-object ratio for each class"):
            with Image.open(mask_path) as mask_image:
                mask = np.array(mask_image)
                total_pixels = mask.size

                pixels_per_class = np.zeros(num_classes)

                for class_id, class_index in class_mapping.items():
                    if background is not None and class_id == background:
                        continue  

                    pixels_per_class[class_index] = np.sum(mask == class_id)

                pixel_ratios = np.zeros(num_classes)
                for class_id, class_index in class_mapping.items():
                    if background is not None and class_id == background:
                        continue  

                    num_objects = objects_per_class[i, class_index]

                    if num_objects > 0:
                        pixel_ratios[class_index] = pixels_per_class[class_index] / (num_objects * total_pixels)
                    else:
                        pixel_ratios[class_index] = 0.0

                all_pixel_ratios.append(pixel_ratios)

        return {
            "distributions": np.array(all_pixel_ratios),
            "num_classes": num_classes,
        }
    
    
    def check_class_distribution(output_dir: str, dataset_name: str, num_classes: int):
        """
        Verifica y visualiza la distribución de clases en los directorios de train, val y test
        según el porcentaje de píxeles que ocupan en las máscaras.
        También guarda los resultados en un CSV y genera un gráfico de barras agrupadas.
        """

        splits = ['train', 'val', 'test']
        distributions = {}
        class_image_counts = {split: {cls: 0 for cls in range(num_classes)} for split in splits}
        total_class_counts = {cls: 0 for cls in range(num_classes)}

        all_files = set()
        for split in splits:
            label_dir = os.path.join(output_dir, split, 'labels')
            mask_files = [os.path.join(label_dir, f) for f in os.listdir(label_dir) if f.endswith('.png')]

            # Check for duplicate files between splits
            for mask_file in mask_files:
                if mask_file in all_files:
                    raise ValueError(f"Duplicate file found across splits: {mask_file}")
                all_files.add(mask_file)

            class_counts = np.zeros(num_classes)
            total_pixels = 0

            for mask_file in mask_files:
                mask = np.array(Image.open(mask_file))
                unique, counts = np.unique(mask, return_counts=True)
                class_counts[unique] += counts
                total_pixels += mask.size

                for cls in unique:
                    class_image_counts[split][cls] += 1
                    total_class_counts[cls] += 1

            class_distribution = class_counts / total_pixels
            distributions[split] = class_distribution

            print(f"Class distribution for {split}:")
            for cls_idx, dist in enumerate(class_distribution):
                print(f"  Class {cls_idx}: {dist:.4f}")
            print("\n")

        # Guardar a CSV
        distributions_clean = {split: dist.tolist() for split, dist in distributions.items()}

        df = pd.DataFrame.from_dict(distributions_clean, orient='index')
        df.columns = [f"Class_{i}" for i in range(num_classes)]
        df.index.name = 'Split'

        # Guardar CSV
        csv_path = os.path.join(output_dir, f"{dataset_name}_pixel_distribution.csv")
        df.to_csv(csv_path,sep=';', encoding="utf8")
        print(f"[CSV] Guardado a {csv_path}")
        
    def calculate_class_distributions(mask_paths: list[str], num_classes: int):
        class_distributions = []
        for mask_path in mask_paths:
            mask = np.array(Image.open(mask_path))
            unique, counts = np.unique(mask, return_counts=True)
            distribution = np.zeros(num_classes)
            distribution[unique] = counts
            class_distributions.append(distribution / distribution.sum())
        
        return np.array(class_distributions)

    def check_class_distribution_with_folds(output_dir: str, dataset_name: str, num_classes: int, n_splits: int):
        """
        Verifica la distribución de clases en los directorios de test y en los folds (train, val) para
        comprobar que la estratificación es correcta.

        Args:
            output_dir: Directorio padre donde se encuentran las carpetas test y folds.
            dataset_name: Nombre del dataset (usado para el nombre del archivo de salida).
            num_classes: Número total de clases en las máscaras.
            n_splits: Número total de folds.
        """

        splits = ['test'] + [f'fold_{i+1}' for i in range(n_splits)]  # Incluye 'test' y los folds
        distributions = {}
        class_image_counts = {split: {cls: 0 for cls in range(num_classes)} for split in splits + [f'fold_{i+1}_train' for i in range(n_splits)] + [f'fold_{i+1}_val' for i in range(n_splits)]}
        total_class_counts = {cls: 0 for cls in range(num_classes)}

        all_files = set()

        # Recorre los splits
        for split in splits:
            if split == 'test':
                # Directorio de test
                label_dir = os.path.join(output_dir, 'test', 'labels')
                mask_files = [os.path.join(label_dir, f) for f in os.listdir(label_dir) if f.endswith('.png')]
                distributions[split] = calculate_class_distributions(mask_files, num_classes)
            else:
                # Directorios de cada fold (train y val)
                label_dir_train = os.path.join(output_dir, split, 'train', 'labels')
                label_dir_val = os.path.join(output_dir, split, 'val', 'labels')
                mask_files_train = [os.path.join(label_dir_train, f) for f in os.listdir(label_dir_train) if f.endswith('.png')]
                mask_files_val = [os.path.join(label_dir_val, f) for f in os.listdir(label_dir_val) if f.endswith('.png')]

                # Crear subdistribuciones para train y val
                distributions[f'{split}_train'] = calculate_class_distributions(mask_files_train, num_classes)
                distributions[f'{split}_val'] = calculate_class_distributions(mask_files_val, num_classes)

        # Mostrar y graficar la distribución
        plot_class_distribution(distributions, num_classes, dataset_name, output_dir)

    def plot_class_distribution(distributions, class_names, num_classes, dataset_name, output_dir):
        bar_width = 0.35
        bar_spacing = 0.2  # Espaciado entre grupos de barras
        group_spacing = 0.6  # Espaciado entre clases

        num_splits = len(distributions)
        total_width = num_splits * bar_width + (num_splits - 1) * bar_spacing

        x_positions = []
        labels = []

        # Crear posiciones y etiquetas para las barras
        for i, split in enumerate(distributions.keys()):
            x_positions.append(np.arange(num_classes) * (total_width + group_spacing) + i * (bar_width + bar_spacing))
            labels.append(split)

        plt.figure(figsize=(16, 8))

        for i, (split, distribution) in enumerate(distributions.items()):
            offset = (i - (num_splits - 1) / 2) * bar_width
            for dist in distribution:
                plt.bar(x_positions[i], distribution, bar_width, label=split)
            

        # Etiquetas y formato
        plt.xlabel('Class')
        plt.ylabel('Pixel proportion')
        plt.title('Class distribution per fold (train/val) and test')
        plt.xticks(np.arange(num_classes) * (total_width + group_spacing) + total_width / 2 - bar_width / 2, class_names)
        plt.legend()
        plt.grid(axis='y', linestyle='--', alpha=0.7)
        plt.tight_layout()

        # Guardar y mostrar la gráfica
        output = os.path.join(output_dir, f"{dataset_name}_class_distribution.png")
        plt.savefig(output, format='png')
        print(f"Plot saved to {output}")

        plt.show()
        plt.close()
    
    def save_split_to_directory(train_df, val_df, test_df, output_dir):
        """
        Saves the partitions into structured directories.
        """
        subsets = {"train": train_df, "val": val_df, "test": test_df}

        for subset, df in subsets.items():
            if df is None:
                continue
            img_dir = os.path.join(output_dir, subset, "images")
            label_dir = os.path.join(output_dir, subset, "labels")
            os.makedirs(img_dir, exist_ok=True)
            os.makedirs(label_dir, exist_ok=True)

            for _, row in df.iterrows():
                shutil.copy(row["img_path"], os.path.join(img_dir, os.path.basename(row["img_path"])))
                shutil.copy(row["mask_path"], os.path.join(label_dir, os.path.basename(row["mask_path"])))

    def check_objects_per_class(output_dir: str, class_names, dataset_name: str, num_classes: int, background: int = None):
        """
        Verifica el número de objetos por clase en las imágenes de los directorios train, val y test
        y genera un diagrama de barras combinando todos los splits en un solo gráfico con colores consistentes para cada clase.

        Args:
            output_dir (str): Directorio donde se encuentran las carpetas train, val y test.
            num_classes (int): Número total de clases en las máscaras.
            background (Optional[int], opcional): Clase de fondo a excluir. Por defecto es None.
        """
        splits = ['train', 'val', 'test']
        objects_per_class_by_split = {}


        # Iterar sobre los splits y calcular los objetos por clase
        for i, split in enumerate(splits):
            label_dir = os.path.join(output_dir, split, 'labels')
            mask_files = [os.path.join(label_dir, f) for f in os.listdir(label_dir) if f.endswith('.png')]

            # Obtener el número de objetos por clase usando calculate_object_number
            objects_per_class = calculate_object_number(mask_files, background)
            
            # Sumar el total de objetos por clase para todas las imágenes en el split
            total_objects = objects_per_class["distributions"].sum(axis=0)

            for j, k in enumerate(total_objects):
                print(f"Total objects from class {j} in split {split}: {k}" )

            # Guardar los resultados por split
            objects_per_class_by_split[split] = total_objects

        df = pd.DataFrame.from_dict(objects_per_class_by_split, orient='index').astype(int)
        df.columns = [f"Class_{i}" for i in range(num_classes)]
        df.index.name = 'Split'

        # Guardar CSV
        csv_path = os.path.join(output_dir, f"{dataset_name}_objects_per_class.csv")
        df.to_csv(csv_path, sep=';')
        print(f"[CSV] Guardado a {csv_path}")

        # csv_path = os.path.join(output_dir, f"{dataset_name}_objects_per_class.csv")
        # df = pd.DataFrame(objects_per_class_by_split)
        # df.index = class_names
        # df.to_csv(csv_path, sep=';')
        # print(f"[CSV] Guardado a {csv_path}")

    def check_ratio_per_class(output_dir: str, class_names, dataset_name: str, background: int = None):
        splits = ['train', 'val', 'test']
        objects_per_class_by_split = {}

        # Iterar sobre los splits y calcular los objetos por clase
        for split in splits:
            label_dir = os.path.join(output_dir, split, 'labels')
            mask_files = [os.path.join(label_dir, f) for f in os.listdir(label_dir) if f.endswith('.png')]

            pixel_ratios = calculate_pixel_ratio(mask_files, background)
            avg_ratios = np.nanmean(pixel_ratios["distributions"], axis=0)
            objects_per_class_by_split[split] = avg_ratios

        # Crear DataFrame con valores absolutos
        df_raw = pd.DataFrame(objects_per_class_by_split)
        df_raw.index = class_names

        df_percent = df_raw * 100

        percent_csv_path = os.path.join(output_dir, f"{dataset_name}_pixel_to_object_ratios_percent.csv")
        df_percent.to_csv(percent_csv_path, sep=';', float_format="%.3f", decimal=',', encoding="utf-8")
        print(f"[CSV] Guardado porcentajes en: {percent_csv_path}")

    def stratify_split(image_files, mask_paths_multilabel,train_fraction, val_fraction, test_fraction, stratification_strategy, background, random_state):
        """
        Stratify-shuffle-split a semantic segmentation dataset into
        train/val/test sets based on pixel-wise class distributions and save
        the results in specified directories.

        Args:
            image_paths: List of file paths to the input images.
            mask_paths: List of file paths to the corresponding segmentation masks.
            train_fraction: Fraction of data to reserve for the training dataset.
            val_fraction: Fraction of data to reserve for the validation dataset.
            test_fraction: Fraction of data to reserve for the test dataset.
            output_dir: Parent directory where the train, val, and test directories will be created. If None, the directories will not be created. Defaults to None.

        Returns:
            Tuple containing three DataFrames: train, val, and test subsets.
            Each DataFrame has two columns: 'img_path' and 'mask_path'.
        """
        image_paths, mask_paths = shuffle(image_files, mask_paths_multilabel, random_state=random_state)

        print("Starting classes stratification...")
        if stratification_strategy.lower() == "pixels":
            result = calculate_pixel_distribution(mask_paths, background)
        elif stratification_strategy.lower() == "objects":
            result = calculate_object_number(mask_paths, background)
        else:
            result = calculate_pixel_ratio(mask_paths, background)
        
        distributions = result["distributions"]
        num_classes = result["num_classes"]

        print(f"Stratification done. {num_classes} classes detected.")
        
        stratifier = IterativeStratification(
            n_splits=2,
            order=1,
            sample_distribution_per_fold=[train_fraction, 1.0 - train_fraction]
        )

        everything_else_indexes, train_indexes = next(stratifier.split(X=np.zeros(len(image_paths)), y=distributions))
        print("Train:", len(train_indexes))
        print("Val+Test:", len(everything_else_indexes))
    
        train_images = [image_paths[i] for i in train_indexes]
        train_masks = [mask_paths[i] for i in train_indexes]

        everything_else_images = [image_paths[i] for i in everything_else_indexes]
        everything_else_masks = [mask_paths[i] for i in everything_else_indexes]

        val_proportion = val_fraction / (val_fraction + test_fraction)  

        stratifier = IterativeStratification(
            n_splits=2,
            order=1,
            sample_distribution_per_fold=[val_proportion, 1.0 - val_proportion]
        )

        test_indexes, val_indexes  = next(stratifier.split(X=np.zeros(len(everything_else_images)), y=distributions[everything_else_indexes]))
        print("Val:", len(val_indexes))
        print("Test:", len(test_indexes))

        val_images = [everything_else_images[i] for i in val_indexes]
        val_masks = [everything_else_masks[i] for i in val_indexes]
        test_images = [everything_else_images[i] for i in test_indexes]
        test_masks = [everything_else_masks[i] for i in test_indexes]

        return train_images, train_masks, val_images, val_masks, test_images, test_masks, num_classes
    
    def random_split(image_files, mask_paths_multilabel, train_fraction, val_fraction, test_fraction, random_state):
        
        train_val_images, test_images, train_val_masks, test_masks = train_test_split(
            image_files,mask_paths_multilabel, test_size=test_fraction, random_state=random_state
        )

        val_fraction_adjusted = val_fraction / (train_fraction + val_fraction)  
        train_images, val_images, train_masks, val_masks = train_test_split(
            train_val_images, train_val_masks, test_size=val_fraction_adjusted, random_state=random_state
        )

        return train_images, train_masks, val_images, val_masks, test_images, test_masks

    def split(image_files, mask_paths_multilabel, num_classes, train_fraction=0.7, val_fraction=0.2, test_fraction=0.1, stratify=True, hold_out=True, stratification_strategy="pixel_objects_ratio", background=None, output_dir=None, random_state=4):
        
        if stratify and stratification_strategy.lower() not in ["pixels", "objects", "pixel_to_object_ratio"]:
            raise ValueError("Invalid value for stratification_strategy. Must be 'pixels', 'objects' or 'pixel_to_object_ratio'.")
        
        if hold_out:
            if not np.isclose(train_fraction + val_fraction + test_fraction, 1.0):
                raise ValueError("The sum of train_fraction, val_fraction, and test_fraction must equal 1.0")

        if mask_paths_multilabel != None and len(image_files) != len(mask_paths_multilabel):
            raise ValueError("The number of images and masks must be the same.")
        
        if hold_out: 
            num_classes = None
            if stratify:
                train_images, train_masks, val_images, val_masks, test_images, test_masks, num_classes = stratify_split(image_files, mask_paths_multilabel, train_fraction, val_fraction, test_fraction, stratification_strategy, background, random_state)
            else:
                train_images, train_masks, val_images, val_masks, test_images, test_masks = random_split(image_files, mask_paths_multilabel, train_fraction, val_fraction, test_fraction, random_state)
        
            train_df = pd.DataFrame({"img_path": train_images, "mask_path": train_masks})
            val_df = pd.DataFrame({"img_path": val_images, "mask_path": val_masks})
            test_df = pd.DataFrame({"img_path": test_images, "mask_path": test_masks})

            save_split_to_directory(train_df, val_df, test_df, output_dir)

            return num_classes
    

    def check_partitions_in_directories(output_dir: str, total_images:int, num_classes: int, test_fraction: float, n_splits: int):
        """
        Verifica la correcta partición y la distribución de clases en los directorios de salida generados por `automate_split_nested`.
        
        Args:
            output_dir: Directorio principal donde se guardan las particiones (test y folds).
            num_classes: Número total de clases en las máscaras.
            test_fraction: Fracción del dataset reservado para el conjunto de test.
            n_splits: Número de folds de validación cruzada.
        """
        # Inicializar los contadores de clases
        class_counts = {split: {cls: 0 for cls in range(num_classes)} for split in ['test'] + [f'fold_{i+1}' for i in range(n_splits)]}

        # 1. Verificar el número de archivos en el directorio 'test'
        test_img_dir = os.path.join(output_dir, 'test', 'images')
        test_label_dir = os.path.join(output_dir, 'test', 'labels')
        
        if not os.path.exists(test_img_dir) or not os.path.exists(test_label_dir):
            raise ValueError(f"Directorio no encontrado: {test_img_dir} o {test_label_dir}")
        
        test_img_files = [f for f in os.listdir(test_img_dir) if f.endswith('.jpg')]
        test_label_files = [f for f in os.listdir(test_label_dir) if f.endswith('.png')]

        if len(test_img_files) != len(test_label_files):
            raise ValueError(f"El número de imágenes y máscaras no coincide en el directorio de 'test'")

        print(f"Test - Imágenes: {len(test_img_files)}, Máscaras: {len(test_label_files)}")

        # 2. Comprobar la distribución de clases en el directorio de test
        for mask_file in test_label_files:
            mask = np.array(Image.open(os.path.join(test_label_dir, mask_file)))
            unique, counts = np.unique(mask, return_counts=True)
            for cls in unique:
                class_counts['test'][cls] += 1

        # 3. Comprobar la distribución de clases en los directorios de los folds
        for fold_idx in range(n_splits):
            fold_img_dir = os.path.join(output_dir, f"fold_{fold_idx + 1}", 'train', 'images')
            fold_label_dir = os.path.join(output_dir, f"fold_{fold_idx + 1}", 'train', 'labels')
            fold_img_dir_val = os.path.join(output_dir, f"fold_{fold_idx + 1}", 'val', 'images')
            fold_label_dir_val = os.path.join(output_dir, f"fold_{fold_idx + 1}", 'val', 'labels')

            
            if not os.path.exists(fold_img_dir) or not os.path.exists(fold_label_dir):
                raise ValueError(f"Directorio no encontrado para el fold {fold_idx + 1}: {fold_img_dir} o {fold_label_dir}")
            
            fold_img_files = [f for f in os.listdir(fold_img_dir) if f.endswith('.jpg')]
            fold_label_files = [f for f in os.listdir(fold_label_dir) if f.endswith('.png')]
            fold_img_dir_val = [f for f in os.listdir(fold_img_dir_val) if f.endswith('.jpg')]
            fold_label_dir_val = [f for f in os.listdir(fold_label_dir_val) if f.endswith('.png')]

            if len(fold_img_files) != len(fold_label_files):
                raise ValueError(f"El número de imágenes y máscaras no coincide en el directorio de 'fold_{fold_idx + 1}'")

            print(f"Fold {fold_idx + 1} - Imágenes (Train): {len(fold_img_files)}, Máscaras (Train): {len(fold_label_files)} | Imágenes (Val): {len(fold_img_dir_val)}, Máscaras (Val): {len(fold_label_dir_val)}")

            # 4. Comprobar la distribución de clases en los folds
            for mask_file in fold_label_files:
                mask = np.array(Image.open(os.path.join(fold_label_dir, mask_file)))
                unique, counts = np.unique(mask, return_counts=True)
                for cls in unique:
                    class_counts[f'fold_{fold_idx + 1}'][cls] += 1

        # 5. Mostrar la distribución de clases en cada conjunto
        print("\nDistribución de clases en 'test' y 'folds':")
        for split in ['test'] + [f'fold_{i+1}' for i in range(n_splits)]:
            print(f"\nDistribución de clases en {split}:")
            for cls in range(num_classes):
                print(f"  Clase {cls}: {class_counts[split][cls]} imágenes")

        # 6. Comprobar la fracción de test
        test_images = len([f for f in os.listdir(os.path.join(output_dir, 'test', 'images')) if f.endswith('.jpg')])

        expected_test_size = total_images * test_fraction
        if not np.isclose(test_images, expected_test_size, atol=1):
            raise ValueError(f"El número de imágenes en el conjunto de test no coincide con la fracción esperada. "
                            f"Esperado: {expected_test_size}, encontrado: {test_images}")
        else:
            print(f"[OK] La fracción de imágenes de test es correcta. Se esperaba: {expected_test_size}, encontrado: {test_images}")

        print("[OK] Todos los directorios y particiones verificadas correctamente.")

    img_path = r"C:\Users\joortif\Desktop\datasets\Preprocesados\forest_freiburg_multiclass\full\images"
    label_path = r"C:\Users\joortif\Desktop\datasets\Preprocesados\forest_freiburg_multiclass\full\labels"

    #img_path = r"C:\Users\joortif\Desktop\datasets\Preprocesados\drone_dataset_reduc\full\images"
    #label_path = r"C:\Users\joortif\Desktop\datasets\Preprocesados\drone_dataset_reduc\full\labels"

    num_classes = 6

    image_files = sorted([os.path.join(img_path, f) for f in os.listdir(img_path) if f.endswith('.jpg')])
    label_files = sorted([os.path.join(label_path, f) for f in os.listdir(label_path) if f.endswith('.png')])

    output_dir = r"C:\Users\joortif\Desktop\Resultados_ImageDatasetAnalyzer\dataset_partitions"

    lbldataset = ImageLabelDataset(img_path, label_path)

    #lbldataset.analyze(plot=False, output=None, verbose=True)

    split(image_files=image_files, mask_paths_multilabel=label_files, num_classes=num_classes, train_fraction=0.7, val_fraction=0.2, test_fraction=0.1, hold_out=True, stratify=False, stratification_strategy="pixel_to_object_ratio", background=None, output_dir=output_dir, random_state=123)
    
    #check_partitions_in_directories(output_dir=output_dir, total_images = 343, num_classes=num_classes, test_fraction=0.1, n_splits=5)

    #check_objects_per_class(output_dir, class_names=['Obstacle', 'Water', 'Nature', 'Water','Moving'], dataset_name="drone", num_classes=5)

    #check_class_distribution(output_dir, "forest_strat", num_classes)

    check_ratio_per_class(output_dir, ['Cielo', 'Obstaculo', 'Vacio', 'Cesped','Vegetacion', 'Camino'], "forest")


import os 
import sys
import logging
import tkinter as tk
from tkinter import filedialog, messagebox
import huggingface_hub
import yaml
from PIL import Image
import numpy as np

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../../")))

from src.datasets.traindataset import TrainingDataset
from src.utils.imagelabelutils import ImageLabelUtils

if __name__ == "__main__":

    
    # image_path = r"C:\Users\joortif\Desktop\datasets\Completos\melanoma_3c\train\images"
    # label_path = r"C:\Users\joortif\Desktop\datasets\Completos\melanoma_3c\train\labels\train_annotations.json"
    # multilabel_path = r"C:\Users\joortif\Desktop\datasets\Preprocesados\melanoma_3c\train\labels"
    # nc = 80
    # background = None
    # output_path = r"C:\Users\joortif\Desktop\datasets\testing_splits\txt"

    # dataset = TrainingDataset(img_path=image_path, multilabel_label_path=multilabel_path, output_path=output_path, label_format="txt", original_label_path=label_path)

    # dataset.split(nc, 0.7, 0.2, 0.1, True, "pixel_distribution", background)

    dict = ImageLabelUtils.get_classes_from_csv(r"C:\Users\joortif\Desktop\datasets\results\results_covid\analysis\metrics.csv", background=255)
    print(dict)

    # # Directorio principal donde se encuentran los folds
    # directorio_base = r"C:\Users\joortif\Desktop\datasets\results_coffee\split"

    # # Lista para almacenar los nombres de archivos en las carpetas val
    # archivos_val_images = set()
    # archivos_val_labels = set()

    # # Variable para marcar si hay archivos repetidos
    # hay_repetidos = False

    # # Recorre los folds
    # for i in range(1, 6):  # Cambia el rango según la cantidad de folds
    #     fold = f'fold_{i}'
    #     fold_path = os.path.join(directorio_base, fold)
        
    #     # Directorios de val/images y val/labels
    #     val_images_path = os.path.join(fold_path, 'val', 'images')
    #     val_labels_path = os.path.join(fold_path, 'val', 'labels')
        
    #     # Comprueba si los directorios existen
    #     if os.path.exists(val_images_path) and os.path.exists(val_labels_path):
    #         # Nombres de archivos en val/images
    #         for archivo in os.listdir(val_images_path):
    #             if archivo.endswith(('.jpg', '.png', '.jpeg')):  # Cambia las extensiones si es necesario
    #                 nombre_archivo = os.path.splitext(archivo)[0]
    #                 if nombre_archivo in archivos_val_images:
    #                     print(f"Archivo repetido en val/images: {archivo}")
    #                     hay_repetidos = True
    #                 archivos_val_images.add(nombre_archivo)
            
    #         # Nombres de archivos en val/labels
    #         for archivo in os.listdir(val_labels_path):
    #             if archivo.endswith(('.txt', '.xml')):  # Cambia las extensiones si es necesario
    #                 nombre_archivo = os.path.splitext(archivo)[0]
    #                 if nombre_archivo in archivos_val_labels:
    #                     print(f"Archivo repetido en val/labels: {archivo}")
    #                     hay_repetidos = True
    #                 archivos_val_labels.add(nombre_archivo)

    # # Resultado final
    # if not hay_repetidos:
    #     print("No hay archivos repetidos en las carpetas val.")
    # else:
    #     print("Se han encontrado archivos repetidos.")









import os
from typing import Optional 
import numpy as np

from segdan.utils.constants import SegmentationType
from segdan.models.smpmodel import SMPModel
from segdan.models.hfstransformermodel import HFTransformerModel

from segdan.datasets.hfdataset import HuggingFaceAdapterDataset
from segdan.datasets.smpdataset import SMPDataset
from torch.utils.data import DataLoader


from segdan.utils.confighandler import ConfigHandler
from segdan.datasets.augments import get_training_augmentation, get_validation_augmentation


smp_models_lower = [name.lower() for name in ConfigHandler.SEMANTIC_SEGMENTATION_MODELS["smp"]]
hf_models_lower = [name.lower() for name in ConfigHandler.SEMANTIC_SEGMENTATION_MODELS["hf"]]

def model_training(model_data: dict, general_data:dict, split_path: str, model_output_path: str, label_path: str, hold_out: bool, classes: Optional[int]):

    epochs = model_data["epochs"]
    batch_size = model_data["batch_size"]
    evaluation_metrics = model_data["evaluation_metrics"]
    selection_metric = model_data["selection_metric"]
    segmentation_type = model_data["segmentation"]
    models = model_data["models"]
    background = general_data["background"]
    
    os.makedirs(model_output_path, exist_ok=True)
    
    if segmentation_type == SegmentationType.SEMANTIC.value:
        models = rename_model_sizes(models)
        semantic_model_training(epochs, batch_size, evaluation_metrics, selection_metric, models, split_path, hold_out, classes, background, model_output_path)

    return

def rename_model_sizes(models: np.ndarray):
    
    for model in models:
        model_size = model["model_size"]
        model_name = model["model_name"]

        if model_name in smp_models_lower:
            model["model_size"] = ConfigHandler.CONFIGURATION_VALUES["model_sizes_smp"].get(model_size)

        if model_name in hf_models_lower:
            model["model_size"] = ConfigHandler.CONFIGURATION_VALUES["model_sizes_hf"].get(model_size)

    return models

def get_data_splits(split_path: str, hold_out: bool, classes: int):
    
    if hold_out:
        train_path = os.path.join(split_path, "train") 
        val_path = os.path.join(split_path, "val") if os.path.exists(os.path.join(split_path, "val")) else None
        test_path = os.path.join(split_path, "test")

        yield {
            "train": SMPDataset(os.path.join(train_path, "images"), os.path.join(train_path, "labels"), classes, get_training_augmentation),
            "val": SMPDataset(os.path.join(val_path, "images"), os.path.join(val_path, "labels"), classes, get_training_augmentation) if val_path else None,
            "test": SMPDataset(os.path.join(test_path, "images"), os.path.join(test_path, "labels"), classes, get_validation_augmentation)
        }
    else:
        fold_dirs = sorted([f for f in os.listdir(split_path) if f.startswith("fold_")])
        for fold in fold_dirs:
            fold_path = os.path.join(split_path, fold)
            train_path = os.path.join(fold_path, "train")
            val_path = os.path.join(fold_path, "val")

            yield {
                "train": SMPDataset(os.path.join(train_path, "images"), os.path.join(train_path, "labels"), classes, get_training_augmentation),
                "val": SMPDataset(os.path.join(val_path, "images"), os.path.join(val_path, "labels"), classes, get_validation_augmentation),
                "test": None
            }

def semantic_model_training(epochs: int, batch_size:int, evaluation_metrics: np.ndarray, selection_metric:str, models: np.ndarray, split_path: str, hold_out: bool, classes: int, background: Optional[int], output_path: str):

    best_metric = -float("inf")
    best_model_path = None

    for fold_idx, data_split in enumerate(get_data_splits(split_path, hold_out, classes)):
        train_dataset = data_split["train"]
        val_dataset = data_split["val"]
        test_dataset = data_split["test"]

        for model_config in models:
            model_size = model_config["model_size"]
            model_name = model_config["model_name"]

            if model_name in hf_models_lower:
                model = HFTransformerModel(model_name, model_size, classes, evaluation_metrics, selection_metric, epochs, batch_size, output_path)

                train_adapter = HuggingFaceAdapterDataset(train_dataset, model.feature_extractor)
                val_adapter = HuggingFaceAdapterDataset(val_dataset, model.feature_extractor) if val_dataset else None
                test_adapter = HuggingFaceAdapterDataset(test_dataset, model.feature_extractor)

                evaluation_metric, candidate_path = model.train(train_adapter, val_adapter, test_adapter)

            else:
                train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=4, persistent_workers=True)
                val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False, num_workers=4, persistent_workers=True) if val_dataset else None
                test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False, num_workers=4, persistent_workers=True) if test_dataset else None

                t_max = epochs * len(train_loader)
                out_classes = len([cls for cls in classes if cls.lower() !="background"])

                model = SMPModel(3, out_classes, evaluation_metrics, selection_metric, epochs, t_max, background, model_name, model_size)

                evaluation_metric, candidate_path = model.train(train_loader, val_loader, test_loader, os.path.join(output_path, "metrics.csv"))

            if evaluation_metric > best_metric:
                print(f"New best model {model_name} found with {selection_metric}={evaluation_metric}")

                if best_model_path and os.path.exists(best_model_path):
                    os.remove(best_model_path)

                best_metric = evaluation_metric
                best_model_path = candidate_path
            else:
                if os.path.exists(candidate_path):
                    os.remove(candidate_path)

    return best_model_path  



from typing import Optional
from segdan.extensions.extensions import LabelExtensions
from segdan.models.depthestimator import DepthEstimator
from segdan.utils.utils import Utils
from segdan.utils.imagelabelutils import ImageLabelUtils
from segdan.converters.converter import Converter

import cv2
import os 
import numpy as np
from tqdm import tqdm

class YOLOToMultilabelConverter(Converter):

    def __init__(self, input_data: str, output_dir: str, img_dir: str, background: Optional[int], depth_model: str ="Intel/dpt-swinv2-tiny-256"):
        super().__init__(input_data, output_dir)
        self.img_dir = img_dir
        self.fill_background = background
        self.depth_model = depth_model
        
    def _read_yolo(self, yolo_path: str):
            annotations = []
            with open(yolo_path) as f:
                for line in f:
                    values = line.strip().split()
                    class_id = int(values[0])
                    values = np.array(list(map(float, values[1:]))).reshape(-1, 2)
                    annotations.append((class_id,values))

            return annotations

    def convert(self):
        
        os.makedirs(self.output_dir, exist_ok=True)

        converted_masks = []
        labels = os.listdir(self.input_data)
        device = Utils.get_device(self.logger)
        depth_estimator = DepthEstimator(self.depth_model, device)

        for label_name in tqdm(labels, desc="Converting labels from TXT YOLO format to multilabel..."):

            label_path = os.path.join(self.input_data, label_name)

            objects = self._read_yolo(label_path)

            image_path = ImageLabelUtils.label_to_image(label_path, self.img_dir, LabelExtensions.JPG.value)

            depth_map = depth_estimator.generate_depth_map(image_path)

            h, w = depth_map.shape
            mask = self._create_empty_mask(h, w, self.fill_background)

            object_depths = []

            for class_id, points in objects:
                mask_obj = np.zeros_like(depth_map, dtype=np.uint8)

                points = self._scale_polygon(points, h, w)

                cv2.fillPoly(mask_obj, [points.astype(np.int32)], 255)
                
                depth_values = depth_map[mask_obj == 255]
                
                if depth_values.size > 0:
                    depth_mean = np.mean(depth_values)
                    object_depths.append((depth_mean, class_id, points))

            object_depths.sort(key=lambda x: x[0], reverse=True)

            for _, class_id, points in object_depths:
                cv2.fillPoly(mask, [points.astype(np.int32)], class_id)

            converted_masks.append(mask)

            ImageLabelUtils.save_multilabel_mask(mask, label_name, self.output_dir)
            
        return converted_masks



import tkinter as tk
from tkinter import ttk


class ReductionFrame(ttk.Frame):
    
    def __init__(self, parent, controller, config_data, final_dict):
        ttk.Frame.__init__(self, parent)

        self.reduction_var = tk.BooleanVar(value=False)
        self.config_data = config_data
        self.controller = controller

        label_title = ttk.Label(self, text="Dataset reduction", font=("Arial", 18, "bold"))
        label_title.grid(row=0, column=0, columnspan=5, pady=(20,10), padx=10)
        
        self.reduction_frame = tk.Frame(self, padx=10, pady=10)
        self.reduction_frame.grid(row=1, column=0, padx=10, pady=10)

        self.grid_rowconfigure(0, weight=0)
        self.grid_columnconfigure(0, weight=1)


        self.row=0
        self.reduction_question_label = tk.Label(self.reduction_frame, text="After grouping images into clusters, we can simplify the dataset by focusing on representative images closest to the cluster center. This reduces redundancy and helps manage large datasets, lowering computational costs and speeding up model training." +
                                                 "It can also guide image selection in Active Learning. In some cases, selecting images farther from the center may be useful to capture outliers or anomalies. Alternatively, randomly selecting images from each cluster combines both strategies for a balanced approach."
                                                                    , wraplength=500)
        self.reduction_question_label.grid(row=self.row, column=0, padx=10, sticky="e") 

        self.row+=1

        self.more_info_reduction = tk.Button(self.reduction_frame, text="More information", command=self.show_more_info)
        self.more_info_reduction.grid(row=self.row, column=0, pady=10)

        self.row+=1

        self.reduction_question_label = tk.Label(self.reduction_frame, text="Do you wish to apply a reduction process to the dataset?", wraplength=500, font=("Arial", 14))
        self.reduction_question_label.grid(row=self.row, column=0, padx=10, pady=(15,10)) 

        self.row+=1
        reduction_button_frame = ttk.Frame(self.reduction_frame)
        reduction_button_frame.grid(row=self.row, column=0, columnspan=5, pady=10)  

        button_reduction_no = tk.Radiobutton(reduction_button_frame, text="No", variable=self.config_data["reduce_images"], value=False)
        button_reduction_no.grid(row=0, column=0, padx=5, pady=5)

        button_reduction_yes = tk.Radiobutton(reduction_button_frame, text="Yes", variable=self.config_data["reduce_images"], value=True)
        button_reduction_yes.grid(row=0, column=1, padx=5, pady=5)

        reduction_button_frame.grid_columnconfigure(0, weight=0)
        reduction_button_frame.grid_columnconfigure(1, weight=0)

        self.row+=1
        button_frame = ttk.Frame(self.reduction_frame)
        button_frame.grid(row=11, column=0, columnspan=5, pady=10, sticky="e")  

        self.reduction_frame.grid_rowconfigure(0, weight=0)
        self.reduction_frame.grid_columnconfigure(0, weight=0)
        self.reduction_frame.grid_columnconfigure(1, weight=0)

        button_back = ttk.Button(button_frame, text="Back", command=self.back)
        button_back.grid(row=0, column=0, padx=50, pady=5, sticky="w")

        button_next = ttk.Button(button_frame, text="Next", command=self.next)
        button_next.grid(row=0, column=1, pady=5, sticky="e")

        button_frame.grid_columnconfigure(0, weight=0)
        button_frame.grid_columnconfigure(1, weight=0)    

    def show_more_info(self):
        tk.messagebox.showinfo("Dataset reduction information", "Once the images have been grouped into different clusters based on their similarity, we can apply a reduction process. This process helps to simplify the dataset, making it easier to work with, especially when dealing with large amounts of data.\n\n" +
                                                                    "One common approach is to focus on the images that are closest to the center of each cluster. These images are typically the most representative of their group, so by selecting just this subset, we can reduce redundant information. " +
                                                                    "This is particularly useful for large datasets that contain many similar images, as it helps to reduce the overall size of the dataset. As a result, this strategy can significantly lower the computational costs when training the segmentation models, " + 
                                                                    "making the process faster and more efficient. Additionally, it can also help in deciding which images to label first in an Active Learning scenario, where labeling is done strategically.\n\n" +
                                                                    "In some cases, however, the images that are farther away from the center of the clusters may be valuable. For example, selecting the most diverse images in the dataset might be helpful for identifying anomalies or outliers." + 
                                                                    "This approach ensures that the dataset includes a variety of different examples, which can be important for detecting unusual patterns.\n\n" +
                                                                    "Finally, another option is to randomly select images from each cluster, combining both of the previous strategies. This can offer a balanced approach, depending on the specific needs of your analysis.")

    def back(self):
        if not self.config_data["cluster_images"]:
            self.controller.show_frame("ClusteringFrame")
            return

        self.controller.show_frame("ClusteringConfigFrame")

    def next(self):
       
        if self.config_data["reduce_images"].get():
            self.controller.show_frame("ReductionConfigFrame")
            return
    
        self.controller.show_frame("DatasetSplitFrame")


import logging
import tkinter as tk
from tkinter import ttk
import sys
import os
import numpy as np



sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../../")))

from imagedatasetanalyzer import ImageDataset

import json

from src.utils.constants import LabelFormat
from src.training.training import model_training
from src.reduction.reduction import reduce_dataset 
from src.clustering.clustering import cluster_images, get_embeddings
from src.converters.converterfactory import ConverterFactory
from src.analysis.analysis import analyze_data
from src.forms.wizard import Wizard
from src.datasets.traindataset import TrainingDataset
from src.training.partition import dataset_split
from src.utils.imagelabelutils import ImageLabelUtils

import warnings
warnings.filterwarnings("ignore", category=UserWarning, module="tensorflow")

def create_logger():
   logger = logging.getLogger("SegmentationPipeline")
   logger.setLevel(logging.INFO)

   if not logger.hasHandlers():
      handler = logging.StreamHandler()
      formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s', datefmt='%H:%M')
      handler.setFormatter(formatter)
      logger.addHandler(handler)
   
   return logger

if __name__ == "__main__":

   def print_config_data(config_data):
        config_json = json.dumps(config_data, indent=4)
        print(config_json)

   app = Wizard()
   app.mainloop()

   config_dict = app.final_dict

   app.destroy()

   if len(config_dict) == 0:
      sys.exit()

   transformer = ConverterFactory()
   logger = create_logger()

   general_data = config_dict["general_data"]
   output_path = general_data["output_path"]
   verbose = general_data["verbose"]
   class_map = general_data["class_mapping"]
   image_path = general_data["image_path"]
   label_path = general_data["label_path"]
   label_format = general_data["label_format"]
   background = general_data["background"]
   
   # Dataset analysis 
   if config_dict["analyze"]:
      analyze_data(general_data, transformer, output_path, class_map, verbose)
      classes = ImageLabelUtils.get_classes_from_csv(os.path.join(output_path, "analysis", "metrics.csv"), background)

   # Embedding clustering
   if config_dict["cluster_images"]:
      clustering_data = config_dict["clustering_data"]
      image_dataset = ImageDataset(image_path)
      embeddings = get_embeddings(clustering_data, image_dataset, verbose, logger)
      clustering_results = cluster_images(clustering_data, image_dataset, embeddings, output_path, verbose, logger)

      # Dataset reduction
      if config_dict["reduce_images"]:
         evaluation_metric = clustering_data["clustering_metric"]
         reduction_data = config_dict["reduction_data"]
         reduction_path = reduce_dataset(reduction_data, clustering_results, evaluation_metric, image_dataset, label_path, 
                                         embeddings, output_path, verbose, logger)
         if reduction_data["use_reduced"]:
            image_path = os.path.join(reduction_path, "images")
            label_path = os.path.join(reduction_path, "labels")

   multilabel_transformation_path = os.path.join(output_path, "transformations", LabelFormat.MASK.value)
   if not os.path.exists(multilabel_transformation_path):
      multilabel_transformation_path = label_path
   
   # Dataset split 
   split_output_path = os.path.join(output_path, "split")
   split_data= config_dict["split_data"]
   dataset = TrainingDataset(image_path, multilabel_transformation_path, split_output_path, label_format, label_path)
   dataset.split(general_data, split_data)
   
   # Model training
   model_data = config_dict["model_data"]
   hold_out = split_data["split_method"]
   model_output_path = os.path.join(output_path, "models")
   model_training(model_data, general_data, split_output_path, model_output_path, multilabel_transformation_path, hold_out=hold_out, 
                  classes=classes)


class ExtensionNotFoundException(Exception):
    def __init__(self, ext):
        self.ext = ext
        self.message = f"Extension '{self.ext}' not found in the supported extensions."
        super().__init__(self.message)


import os
from torch.utils.data import Dataset as BaseDataset
import os
import cv2
import numpy as np

from segdan.utils.confighandler import ConfigHandler

class SMPDataset(BaseDataset):
    
    def __init__(self, images_dir, masks_dir, classes, augmentation=None, background=None):
        self.images_fps = []
        self.ids = []
        for fname in os.listdir(images_dir):
            if fname.lower().endswith(tuple(ConfigHandler.VALID_IMAGE_EXTENSIONS)):
                fpath = os.path.join(images_dir, fname)
                img = cv2.imread(fpath)
                if img is not None:
                    self.images_fps.append(fpath)
                    self.ids.append(fname)
        
        masks_ids = [os.path.splitext(image_id)[0]+'.png' for image_id in self.ids]
        self.masks_fps = [os.path.join(masks_dir, image_id) for image_id in masks_ids]
        self.background_class = background
        
        self.classes = classes
        if classes:
            self.class_values = [self.classes.index(cls.lower()) for cls in classes]
        else:
            self.class_values = list(range(len(self.classes)))
            
        # Create a remapping dictionary: class value in dataset -> new index (0, 1, 2, ...)
        # Background will always be 255, other classes will be remapped starting from 1.
        self.class_map = {self.background_class: 255} if self.background_class is not None else {}
        self.class_map.update(
            {
                v: i
                for i, v in enumerate(self.class_values)
                if v != self.background_class
            }
        )
                
        self.augmentation = augmentation

    def __getitem__(self, i):
        # Read the image
        image = cv2.imread(self.images_fps[i])
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)  # Convert BGR to RGB

        # Read the mask in grayscale mode
        mask = cv2.imread(self.masks_fps[i], 0)
        
        # Create a blank mask to remap the class values
        mask_remap = np.full_like(mask, 255, dtype=np.uint8)

        # Remap the mask according to the dynamically created class map
        for class_value, new_value in self.class_map.items():
            mask_remap[mask == class_value] = new_value
        
        unique_vals = np.unique(mask)

        valid_values = set(self.class_map.values()) | {255}
        assert set(np.unique(mask_remap)).issubset(valid_values), f"Unexpected values in remapped mask: {np.unique(mask_remap)}"

        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask_remap)
            image, mask_remap = sample["image"], sample["mask"]
            
            unique_vals = np.unique(mask_remap)
            assert any(val != 255 for val in unique_vals), f"Mask at index {i} became empty after augmentation: {unique_vals}"

        image = image.transpose(2, 0, 1)
        return image, mask_remap

    def __len__(self):
        return len(self.ids)
    
    def get_class_map(self):
        return self.class_map


import os
import sys
import cv2
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
from tqdm import tqdm


sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../../")))

from converters.yolo_to_multilabel import YOLOToMultilabelConverter
from converters.json_to_multilabel import JSONToMultilabelConverter
from converters.multilabel_to_instance_seg import MultilabelToInstanceSegmentationConverter


if __name__ == "__main__":

    def overlay_mask_on_image(image_path, mask_path, output_path=None):
        """
        Superpone la máscara de segmentación sobre la imagen original y muestra el resultado.
        
        :param image_path: Ruta de la imagen original.
        :param mask_path: Ruta de la máscara de segmentación.
        :param output_path: Ruta donde guardar la imagen superpuesta (opcional).
        """
        # Cargar la imagen original con PIL
        with Image.open(image_path) as image:
            image = Image.open(image_path).convert('RGB')
            
            # Cargar la máscara (en escala de grises)
            mask = Image.open(mask_path).convert('L')
            mask = np.array(mask)

            # Convertir la imagen a un array de numpy para manipularla
            image_np = np.array(image)

            # Crear una máscara de color (verde en este caso) para los píxeles no cero
            colored_mask = np.zeros_like(image_np)  # Crear una máscara del mismo tamaño que la imagen
            colored_mask[mask == 0] = [0, 255, 0]  # Asignar el color verde a los píxeles no cero en la máscara
            colored_mask[mask == 1] = [255, 0, 0]
            colored_mask[mask == 2] = [0, 0, 255]
            colored_mask[mask == 3] = [0, 0, 125]
            # Superponer la imagen original y la máscara con algo de transparencia
            overlay = np.where(colored_mask == 0, image_np, (0.6 * colored_mask + 0.4 * image_np).astype(np.uint8))

            # Mostrar la imagen original con la máscara superpuesta
            plt.figure(figsize=(10, 10))
            plt.imshow(overlay)
            plt.axis('off')  # Ocultar los ejes
            #plt.show()
            plt.close()

            # Si se especificó un `output_path`, guardar la imagen superpuesta
            if output_path:
                # Guardar la imagen superpuesta usando PIL
                overlay_image = Image.fromarray(overlay)
                overlay_image.save(output_path)
                #print(f"✅ Imagen superpuesta guardada en: {output_path}")


    transformer = JSONToMultilabelTransformer()

    img_path = r"C:\Users\joortif\Desktop\datasets\Completos\melanoma_3c\train\images"
    lbl_path = r"C:\Users\joortif\Desktop\datasets\Completos\melanoma_3c\train\labels"
    output_dir = r"C:\Users\joortif\Desktop\Resultados_SegDAN\transformers\YoloToMultilabel"
    output_superposition = r"C:\Users\joortif\Desktop\Resultados_SegDAN\transformers\superposiciones_c3depth"

    res = transformer.transform(lbl_path, img_path, 3, output_dir=output_dir, verbose=True)

    saved_counts = {0: 0, 1: 0, 2: 0}

    images = os.listdir(img_path)

    for i, image in enumerate(images):
        image_name, _ = os.path.splitext(image)
        mask_name = f"{image_name}.png"
        mask_path = os.path.join(output_dir, mask_name)

        # Verificar si la máscara existe
        if not os.path.exists(mask_path):
            continue
        
        # Cargar la máscara para identificar qué clases contiene
        mask = np.array(Image.open(mask_path).convert("L"))

        # Identificar las clases presentes en la máscara
        unique_classes = np.unique(mask)

        for class_id in unique_classes:
            if class_id in saved_counts and saved_counts[class_id] < 10:
                output_path = os.path.join(output_superposition, f"superposicion_{class_id}_{saved_counts[class_id]}.png")
                overlay_mask_on_image(os.path.join(img_path, image), mask_path, output_path)
                saved_counts[class_id] += 1

            # Si ya tenemos 10 imágenes de cada clase, terminar el proceso
            if all(count >= 10 for count in saved_counts.values()):
                print("✅ Se han guardado 10 imágenes por clase. Proceso terminado.")
                exit()
    
    # YOLO TO MULTILABEL
    """transformer = YOLOToMultilabelTransformer()

    objects = transformer.transform(lbl_path, img_path, '.jpg', 80, output_path=output_dir, verbose=True)"""

    """
    # MULTILABEL TO MULTICHANNEL
    
    batch_masks = np.array([
    [[0, 1, 1, 1], 
     [1, 0, 2, 0], 
     [0, 3, 2, 0]],  # Primera imagen
    [[3, 2, 3, 3], 
     [2, 1, 2, 0], 
     [4, 0, 5, 0]]   # Segunda imagen
    ])

    transformer = MultilabelToInstanceSegmentationTransformer()

    # Convertir el batch a formato multicanal sin la clase 0
    multichannel_batch = transformer.transform(batch_masks, background=0)


    for i, instance_mask in enumerate(multichannel_batch):
        print(f"Instancia {i + 1} pertenece a la clase {np.unique(instance_mask[instance_mask != 0])}:")
        print(instance_mask)
    print(multichannel_batch.shape)"""

    


import numpy as np

from segdan.utils.utils import Utils
from segdan.metrics.clusteringmetrics import get_scoring_function
from segdan.utils.constants import ClusteringModelName

class ClusteringModel():

    def __init__(self, model, args, embeddings, metric, visualization_technique, plot, output_path):
        self.model = model
        self.args = args
        self.plot = plot
        self.output_path = output_path
        self.embeddings = embeddings
        self.metric = metric
        self.visualization_technique = visualization_technique


    def _clustering(self, scoring_function, params):
        if self.plot:
            labels = self.model.clustering(**params, reduction = self.visualization_technique, output=self.output_path)
        else:
            labels = self.model.clustering(**params)
        return scoring_function(self.embeddings, labels), labels
    
    def _plot_best_model(self, best_labels, model_name):
        num_clusters = len(np.unique(best_labels))
        embeddings_2d = self.model.reduce_dimensions(self.visualization_technique)
        self.model.plot_clusters(embeddings_2d, best_labels, num_clusters, self.visualization_technique, model_name, self.output_path)
    
    def get_param(self, param_name):
        
        if f"{param_name}_range" in self.args:
            return Utils.params_to_range(self.args[f"{param_name}_range"])
        return self.args.get(param_name)  
    
    def get_param(self, param_name):

        if "linkages" in self.args and param_name.strip() == "linkage":
            return self.args.get("linkages")

        if "_range" in param_name.strip():
            return Utils.params_to_range(self.args[f"{param_name}"])
        
        return self.args.get(param_name) 

    def train(self, model_name, verbose: bool):

        print(f"Using {model_name} model for clustering image embeddings...")

        scoring_function = get_scoring_function(self.metric)

        params = {param_name: self.get_param(param_name) for param_name in self.args.keys() if param_name != 'random_state'}

        if all(not "_range" in param for param in params):

            score, labels = self._clustering(scoring_function, params)

            if model_name == 'kmeans':

                params["random_state"] = self.args["random_state"]

            return (*params.values(), score, labels)

        if model_name == ClusteringModelName.KMEANS.value:
            best_n_clusters, random_state, best_score, best_labels =  self.model.find_best_n_clusters(params["n_clusters_range"], self.metric, self.plot, self.output_path)

            self._plot_best_model(best_labels, model_name)

            return best_n_clusters, random_state, best_score, best_labels
        
        if model_name == ClusteringModelName.AGGLOMERATIVE.value:

            best_k, best_linkage, best_score, best_labels = self.model.find_best_agglomerative_clustering(params["n_clusters_range"], self.metric, params["linkages"], self.plot, self.output_path)

            self._plot_best_model(best_labels, model_name)

            return best_k, best_linkage, best_score, best_labels
        
        if model_name == ClusteringModelName.DBSCAN.value:

            best_eps, best_min_samples, best_score, best_labels = self.model.find_best_DBSCAN(params["eps_range"], params["min_samples_range"], self.metric, self.plot, self.output_path, verbose)

            self._plot_best_model(best_labels, model_name)
            
            return best_eps, best_min_samples, best_score, best_labels
        if model_name == ClusteringModelName.OPTICS.value:

            best_min_samples, best_score, best_labels = self.model.find_best_OPTICS(params["min_samples_range"], self.metric, self.plot, self.output_path, verbose)

            self._plot_best_model(best_labels, model_name)

            return best_min_samples, best_score, best_labels


from transformers import pipeline
from PIL import Image
import cv2
import numpy as np

class DepthEstimator:
    def __init__(self, depth_model="Intel/dpt-large", device=0):
        self.pipe = pipeline(task="depth-estimation", model=depth_model, device=device)

    def generate_depth_map(self, image_path):
        img = Image.open(image_path)

        result = self.pipe(img)
        depth_map = result["depth"]

        depth_map = np.array(depth_map)

        depth_map = cv2.normalize(depth_map, None, 0, 255, cv2.NORM_MINMAX)
        depth_map = np.uint8(depth_map)

        w, h = img.size
        depth_map_resized = cv2.resize(depth_map, (w, h), interpolation=cv2.INTER_LINEAR)

        return depth_map_resized


import tkinter as tk
from tkinter import ttk

class ClusteringFrame(ttk.Frame):

    def __init__(self, parent, controller, config_data, final_dict):
        ttk.Frame.__init__(self, parent)

        self.config_data = config_data
        self.controller = controller

        label_title = ttk.Label(self, text="Image similarity", font=("Arial", 18, "bold"))
        label_title.grid(row=0, column=0, columnspan=5, pady=(20,10), padx=10)
        
        self.clustering_frame = tk.Frame(self, padx=10, pady=10)
        self.clustering_frame.grid(row=1, column=0, padx=10, pady=10)

        self.grid_rowconfigure(0, weight=0)
        self.grid_columnconfigure(0, weight=1)

        self.row=0
        self.clustering_question_label = tk.Label(self.clustering_frame, text="Image embedding models allow us to analyze image similarity by converting images into numerical representations and using clustering algorithms like KMeans, " +
                                                  "Agglomerative Clustering, DBSCAN, and OPTICS to group similar images. You can use preset or optimized settings for these methods. If this step isn’t necessary, you can skip it to speed up the process.", wraplength=500)
        self.clustering_question_label.grid(row=self.row, column=0, padx=10, sticky="e") 

        self.row+=1

        self.more_info_button = tk.Button(self.clustering_frame, text="More information", command=self.show_more_info)
        self.more_info_button.grid(row=self.row, column=0, pady=10)

        self.row+=1

        self.clustering_question_label = tk.Label(self.clustering_frame, text="Do you wish to apply clustering model?", wraplength=500, font=("Arial", 14))
        self.clustering_question_label.grid(row=self.row, column=0, padx=10, pady=(15,10)) 

        self.row+=1
        clustering_button_frame = ttk.Frame(self.clustering_frame)
        clustering_button_frame.grid(row=self.row, column=0, columnspan=5, pady=10)  

        button_clustering_no = tk.Radiobutton(clustering_button_frame, text="No", variable=self.config_data["cluster_images"], value=False)
        button_clustering_no.grid(row=0, column=0, padx=5, pady=5)

        button_clustering_yes = tk.Radiobutton(clustering_button_frame, text="Yes", variable=self.config_data["cluster_images"], value=True)
        button_clustering_yes.grid(row=0, column=1, padx=5, pady=5)

        clustering_button_frame.grid_columnconfigure(0, weight=0)
        clustering_button_frame.grid_columnconfigure(1, weight=0)

        self.row+=1
        button_frame = ttk.Frame(self.clustering_frame)
        button_frame.grid(row=11, column=0, columnspan=5, pady=10, sticky="e")  

        self.clustering_frame.grid_rowconfigure(0, weight=0)
        self.clustering_frame.grid_columnconfigure(0, weight=0)
        self.clustering_frame.grid_columnconfigure(1, weight=0)

        button_back = ttk.Button(button_frame, text="Back", command=lambda: self.controller.show_frame("AnalysisConfigFrame"))
        button_back.grid(row=0, column=0, padx=50, pady=5, sticky="w")

        button_next = ttk.Button(button_frame, text="Next", command=self.next)
        button_next.grid(row=0, column=1, pady=5, sticky="e")

        button_frame.grid_columnconfigure(0, weight=0)
        button_frame.grid_columnconfigure(1, weight=0)    

    def show_more_info(self):

        tk.messagebox.showinfo("Image similarity information", "Using image embedding models, we can analyze the similarity between images in a dataset in two steps.\n\n"+
                                                "First, an image embedding model (chosen by the user) converts images into numerical representations, capturing their most important features " +
                                                "while reducing unnecessary details. This allows us to compare images more effectively.\n\n" +
                                                "Next, clustering algorithms group similar images together based on these numerical representations. You can choose from different clustering " +
                                                "methods, such as KMeans, Agglomerative Clustering, DBSCAN, and OPTICS, either with preset settings or optimized through a search process.\n\n" +
                                                "If this step is not relevant to your needs, you can skip it to speed up the application process.")

    def next(self):

        if self.config_data["cluster_images"].get(): 
            self.controller.show_frame("ClusteringConfigFrame")
            return
        
        self.controller.show_frame("DatasetSplitFrame")


from transformers import Trainer

import torch.nn.functional as F
import torch
import numpy as np

import segmentation_models_pytorch as smp

from segdan.metrics.custom_metric import custom_metric
from segdan.metrics.segmentationmetrics import dice_score

class HFSegmentationTrainer(Trainer):
    def __init__(self, *args, num_classes=None, ignore_index=None, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.ignore_index = ignore_index
        self.num_classes = num_classes
        self.binary = num_classes == 1

    def compute_loss(self, model, inputs, return_outputs=False, **kwargs):
        outputs = model(**inputs)  

        loss = outputs.loss  
        
        return (loss, outputs) if return_outputs else loss
    
    def compute_metrics(eval_pred, num_classes, stage="train", ignore_index=255):
        predictions, label_ids = eval_pred
        
        class_logits = torch.from_numpy(predictions[0])      
        masks_q = torch.from_numpy(predictions[1])     
        
        probs = torch.softmax(class_logits, dim=-1)[..., 1]
        raw_mask = (probs[:, :, None, None] * masks_q).sum(dim=1)
        
        masks_tensor = torch.tensor(np.array(label_ids[0])).bool()  # (B, M, H, W)
        classes_tensor = torch.tensor(label_ids[1])

        gt_masks = []
        pred_masks = []

        for i in range(masks_tensor.shape[0]):
            gt_mask = torch.zeros_like(masks_tensor[i, 0])
            for m, c in zip(masks_tensor[i], classes_tensor[i]):
                if c.item() == 1:
                    gt_mask |= m
            gt_masks.append(gt_mask)

            pred = F.interpolate(
                raw_mask[i].unsqueeze(0).unsqueeze(0),  # (1, 1, H, W)
                size=gt_mask.shape,
                mode="bilinear",
                align_corners=False
            ).squeeze()
            pred_masks.append((pred.sigmoid() > 0.5).bool())
            
        pred_masks_array = torch.stack(pred_masks)
        gt_masks_array = torch.stack(gt_masks)

        tp, fp, fn, tn = smp.metrics.get_stats(pred_masks_array, gt_masks_array, mode="binary")

        dataset_iou = smp.metrics.iou_score(tp, fp, fn, tn, reduction="micro")
        iou_per_class = smp.metrics.iou_score(tp, fp, fn, tn, reduction="none")
        dataset_accuracy = smp.metrics.accuracy(tp, fp, fn, tn, reduction="micro")
        dataset_precision = smp.metrics.precision(tp, fp, fn, tn, reduction="micro")
        dataset_recall = smp.metrics.recall(tp, fp, fn, tn, reduction="micro")
        dataset_f1_score = smp.metrics.f1_score(tp, fp, fn, tn, reduction="micro")
        dataset_dice = custom_metric(tp, fp, fn, tn, dice_score, reduction="micro") 
        dice_per_class = custom_metric(tp, fp, fn, tn, dice_score, reduction="none") 

        metrics = {
            f"Accuracy": dataset_accuracy,
            f"Precision": dataset_precision,
            f"Recall": dataset_recall,
            f"F1_score": dataset_f1_score,
            f"Dice": dataset_dice,
            f"Iou": dataset_iou,
        }
        
        for class_idx in range(num_classes):
            iou_class = iou_per_class[class_idx]
            metrics[f"Class_{class_idx}_iou"] = iou_class.mean().item()  

            dice_class = dice_per_class[class_idx]
            metrics[f"Class_{class_idx}_dice"] = dice_class.mean().item()
        
        return metrics


import tkinter as tk
from tkinter import ttk, font
import huggingface_hub
from tktooltip import ToolTip
import webbrowser

from segdan.utils.confighandler import ConfigHandler
from segdan.forms.formutils import FormUtils
from segdan.forms.clusteringmodelform import ClusteringModelForm
from segdan.utils.constants import Framework

import torchvision.models as models
import tensorflow as tf


class ClusteringConfigFrame(ttk.Frame):
    def __init__(self, parent, controller, config_data, final_dict):
            ttk.Frame.__init__(self, parent)
            self.top = parent
            self.config_data = config_data

            if "clustering_data" not in self.config_data:
                self.config_data["clustering_data"] = {
                    "embedding_model": {
                        "framework": self.config_data.get("framework", tk.StringVar(value=Framework.OPENCV.value)),
                        "name": self.config_data.get("name", tk.StringVar(value="")),
                        "name_other": self.config_data.get("name_other", tk.StringVar(value="")),
                        "resize_height": self.config_data.get("resize_height", tk.StringVar(value="224")),
                        "resize_width": self.config_data.get("resize_width", tk.StringVar(value="224")),
                        "embedding_batch_size": self.config_data.get("embedding_batch_size", tk.StringVar(value="8")), 
                        "lbp_radius": self.config_data.get("lbp_radius", tk.StringVar(value="16")),
                        "lbp_num_points": self.config_data.get("lbp_num_points", tk.StringVar(value="48")),
                        "lbp_method": self.config_data.get("lbp_method", tk.StringVar(value="uniform")),
                    },
                    "clustering_models": self.config_data.get("clustering_models", {}),
                    "clustering_metric": self.config_data.get("clustering_metric", tk.StringVar(value="calinski")),
                    "plot": self.config_data.get("plot", tk.BooleanVar(value=False)),
                    "visualization_technique": self.config_data.get("visualization_technique", tk.StringVar(value="pca")),
                }
            self.clustering_data = self.config_data["clustering_data"]
            self.controller = controller

            self.grid_rowconfigure(0, weight=0)
            self.grid_columnconfigure(0, weight=1)

            label_title = ttk.Label(self, text="Image embedding and clustering", font=("Arial", 18, "bold"))
            label_title.grid(row=0, column=0, columnspan=5, pady=(20,10), padx=10)
            
            self.clustering_frame = tk.Frame(self, padx=10, pady=10)
            self.clustering_frame.grid(row=1, column=0, padx=10, pady=10)

            self.clustering_config_widgets()

            button_frame = ttk.Frame(self.clustering_frame)
            button_frame.grid(row=9, column=0, columnspan=5, pady=10, sticky="e")  

            self.clustering_frame.grid_rowconfigure(0, weight=0)
            self.clustering_frame.grid_columnconfigure(0, weight=0)
            self.clustering_frame.grid_columnconfigure(1, weight=0)


            button_back = ttk.Button(button_frame, text="Back", command=lambda: controller.show_frame("ClusteringFrame"))
            button_back.grid(row=0, column=0, padx=50, pady=5, sticky="w")

            button_next = ttk.Button(button_frame, text="Next", command=self.next)
            button_next.grid(row=0, column=1, pady=5, sticky="e")

            button_frame.grid_columnconfigure(0, weight=1)
            button_frame.grid_columnconfigure(1, weight=1)

    def clustering_config_widgets(self):
        frameworks = ConfigHandler.CONFIGURATION_VALUES['frameworks']

        lbp_methods = ConfigHandler.CONFIGURATION_VALUES['lbp_methods']
        clustering_metrics = ConfigHandler.CONFIGURATION_VALUES['clustering_metric']
        visualization_techniques = ConfigHandler.CONFIGURATION_VALUES['visualization_techniques']

        vcmd = (self.top.register(self.validate_numeric), "%P")

        self.embedding_model_labelframe = ttk.LabelFrame(self.clustering_frame, text="Embedding model configuration", padding=(20, 10))

        self.framework_label = tk.Label(self.embedding_model_labelframe, text="Framework *")
        self.framework_label.grid(row=0, column=0, padx=5)
        ToolTip(self.framework_label, msg="Select the framework used for image embeddings.")
        
        self.name_label = tk.Label(self.embedding_model_labelframe, text="Name *")
        self.name_combobox = ttk.Combobox(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["name"], state="readonly")

        self.lbp_name_entry = tk.Entry(self.embedding_model_labelframe, textvariable=tk.StringVar(value="Local Binary Pattern (LBP)"), state="readonly", width=25)
        ToolTip(self.name_label, msg="Embedding model name.")
        self.framework_combobox = ttk.Combobox(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["framework"], values=frameworks, state="readonly", width=15)

        self.other_name_label = tk.Label(self.embedding_model_labelframe, text="Custom model name *")
        self.other_name_entry = tk.Entry(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["name_other"])
        ToolTip(self.other_name_label, msg="Custom embedding model name.")

        self.custom_model_info_label = tk.Label(self.embedding_model_labelframe)

        self.resize_height_label = tk.Label(self.embedding_model_labelframe, text="Resize height *")  
        self.resize_height_entry = tk.Entry(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["resize_height"], width=10, validate="key", validatecommand=vcmd)
        ToolTip(self.resize_height_label, msg="Image resizing height before applying the embedding model.")

        self.resize_width_label = tk.Label(self.embedding_model_labelframe, text="Resize width *")        
        self.resize_width_entry = tk.Entry(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["resize_width"], width=10, validate="key", validatecommand=vcmd)
        ToolTip(self.resize_width_label, msg="Image resizing width before applying the embedding model.")

        self.batch_size_label = tk.Label(self.embedding_model_labelframe, text="Batch size *")  
        self.batch_size_entry = tk.Entry(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["embedding_batch_size"], width=10, validate="key", validatecommand=vcmd)
        ToolTip(self.batch_size_label, msg="Defines the number of samples per training step. Larger values need more memory.")

        self.lbp_radius_label = tk.Label(self.embedding_model_labelframe, text="Radius *")        
        self.lbp_radius_entry = tk.Entry(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["lbp_radius"], width=10, validate="key", validatecommand=vcmd)
        ToolTip(self.lbp_radius_label, msg="Defines the neighborhood radius for LBP feature extraction.\nHigher values capture texture at a larger scale.")

        self.lbp_num_points_label = tk.Label(self.embedding_model_labelframe, text="Num Points *")       
        self.lbp_num_points_entry = tk.Entry(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["lbp_num_points"], width=10, validate="key", validatecommand=vcmd)
        ToolTip(self.lbp_num_points_label, msg="Number of sampling points around the defined radius.\nMore points capture finer texture details.")

        self.lbp_method_label = tk.Label(self.embedding_model_labelframe, text="Method *")
        self.lbp_method_combobox = ttk.Combobox(self.embedding_model_labelframe, textvariable=self.clustering_data['embedding_model']["lbp_method"], values=lbp_methods, state="readonly",  width=15)
        ToolTip(self.lbp_method_label, msg="LBP computation method.\nEach method affects how texture patterns are generated.")

        self.embedding_model_labelframe.grid(row=0, column=0, padx=5, pady=10, columnspan=4, sticky="ew")

        self.clustering_labelframe = ttk.LabelFrame(self.clustering_frame, text="Clustering configuration", padding=(20, 10))
        self.clustering_labelframe.grid(row=1, column=0, padx=5, pady=10, columnspan=4, sticky="ew")

        self.clustering_models_text = tk.Text(self.clustering_labelframe, height=5, width=15, state="disabled")


        self.clustering_models_label = tk.Label(self.clustering_labelframe, text="Clustering models *")
        self.clustering_models_label.grid(row=0, column=0, padx=10, pady=5)
        ToolTip(self.clustering_models_label, msg="Clustering models list to use with the embeddings.")

        self.clustering_metric_label = tk.Label(self.clustering_labelframe, text="Clustering metric *")
        self.clustering_metric_label.grid(row=0, column=2, padx=10, pady=5)

        ToolTip(self.clustering_metric_label, msg="Metric that defines how similarity between image clusters is measured.")
        
        self.clustering_models_text.grid(row=1, column=0, padx=10, pady=5, rowspan=2)
        
        self.add_model_bt = tk.Button(self.clustering_labelframe, text="Add model", command=lambda: self.open_model_form())
        self.add_model_bt.grid(row=1, column=1)
        ToolTip(self.add_model_bt, msg="Add a new clustering model.")
        ttk.Combobox(self.clustering_labelframe, textvariable=self.clustering_data["clustering_metric"], values=clustering_metrics, state="readonly", width=10).grid(row=1, column=2)        

        self.edit_model_bt = tk.Button(self.clustering_labelframe, text="Edit model", command=lambda: self.open_model_form(edit=True))
        self.edit_model_bt.grid(row=2, column=1)
        ToolTip(self.edit_model_bt, msg="Edit an existing clustering model.")

        self.edit_model_bt.grid_remove()

        self.visualization_labelframe = ttk.LabelFrame(self.clustering_frame, text="Visualization parameters", padding=(20, 10))
        self.visualization_labelframe.grid(row=2, column=0, padx=5, pady=10, columnspan=4, sticky="ew")

        self.visualization_label = tk.Label(self.visualization_labelframe, text="Visualization technique *")
        self.visualization_technique_combobox = ttk.Combobox(self.visualization_labelframe, textvariable=self.clustering_data["visualization_technique"], values=visualization_techniques, state="readonly", width=6)        
        ToolTip(self.visualization_label, msg="Dimensionality reduction technique applied to the embeddings for plotting in 2D.")

        self.visualization_label.grid(row=1, column=1)


        self.clustering_plot_label = tk.Label(self.visualization_labelframe, text="Plot")
        self.clustering_plot_label.grid(row=0, column=0, padx=10, pady=5)
        ToolTip(self.clustering_plot_label, msg="Enables visualization of clustering results by reducing the dimensionality of the embeddings.\nPlots images in a 2 dimensional space.")

        ttk.Checkbutton(self.visualization_labelframe, variable=self.clustering_data["plot"], command=lambda: FormUtils.toggle_label_entry(self.clustering_data["plot"], self.visualization_label, 
                                                                                                                          self.visualization_technique_combobox, None, 0, 1)).grid(row=1, column=0, padx=5, pady=5)
        self.visualization_technique_combobox.grid(row=0, column=1, padx=10, pady=5)

        

        self.framework_combobox.bind("<<ComboboxSelected>>", self.update_framework_fields)
        self.name_combobox.bind("<<ComboboxSelected>>", self.select_other_model)

        self.name_label.grid_forget()
        self.other_name_label.grid_forget()
        self.custom_model_info_label.grid_forget()
        self.other_name_entry.grid_forget()
        self.name_combobox.grid_forget()
        self.visualization_label.grid_forget()
        self.visualization_technique_combobox.grid_forget()

        self.framework_combobox.grid(row=1, column=0, padx=10, pady=5)
        self.name_label.grid(row=0, column=1, padx=10, pady=5)
        self.lbp_name_entry.grid(row=1, column=1, padx=10, pady=5)
        self.resize_height_label.grid(row=0, column=2, padx=10)
        self.resize_height_entry.grid(row=1, column=2, padx=10, pady=5)
        self.resize_width_label.grid(row=0, column=3, padx=10)
        self.resize_width_entry.grid(row=1, column=3, padx=10, pady=5)
        self.lbp_radius_label.grid(row=2, column=1, padx=10)
        self.lbp_radius_entry.grid(row=3, column=1, padx=10, pady=5)
        self.lbp_num_points_label.grid(row=2, column=2, padx=10)
        self.lbp_num_points_entry.grid(row=3, column=2, padx=10, pady=5)
        self.lbp_method_label.grid(row=2, column=3, padx=10)
        self.lbp_method_combobox.grid(row=3, column=3, padx=10, pady=5)
        self.batch_size_label.grid(row=2, column=0, padx=10, pady=5)
        self.batch_size_entry.grid(row=3, column=0, padx=10, pady=5)

    def select_other_model(self, event):

        selected_value = self.name_combobox.get()
        framework = self.framework_combobox.get()

        if selected_value.lower() == "other":
            self.other_name_entry.delete(0, tk.END)  
            f = font.Font(self.custom_model_info_label, self.custom_model_info_label.cget("font"))
            f.configure(underline=True)
            self.custom_model_info_label.configure(font=f)

            if framework == Framework.HUGGINGFACE.value:
                self.other_name_label.grid(row=0, column=3, pady=5, padx=5)
                self.other_name_entry.grid(row=1, column=3, pady=5,padx=5)

                self.custom_model_info_label.config(text = "HuggingFace models for image feature extraction", fg="blue", cursor="hand2", wraplength=200)
                self.custom_model_info_label.bind("<Button-1>", lambda e: webbrowser.open(ConfigHandler.FRAMEWORK_URLS["huggingface"]))
                self.custom_model_info_label.grid(row=2, column=3, pady=5, padx=5)

                return
            
            self.other_name_label.grid(row=0, column=5, pady=5, padx=5)
            self.other_name_entry.grid(row=1, column=5, pady=5,padx=5)

            if framework == Framework.PYTORCH.value:
                self.custom_model_info_label.config(text = "Pytorch models for image feature extraction", fg="blue", cursor="hand2", wraplength=200)
                self.custom_model_info_label.bind("<Button-1>", lambda e: webbrowser.open(ConfigHandler.FRAMEWORK_URLS["pytorch"]))
                self.custom_model_info_label.grid(row=2, column=5, pady=5, padx=5)
                return

            self.custom_model_info_label.config(text = "TensorFlow models for image feature extraction", fg="blue", cursor="hand2", wraplength=200)
            self.custom_model_info_label.bind("<Button-1>", lambda e: webbrowser.open(ConfigHandler.FRAMEWORK_URLS["tensorflow"]))
            self.custom_model_info_label.grid(row=2, column=5, pady=5, padx=5)
            return
        
        self.other_name_entry.grid_forget()
        self.other_name_label.grid_forget()
        self.custom_model_info_label.grid_forget()

    def update_framework_fields(self, event):
        framework = self.clustering_data['embedding_model']["framework"].get()
        self.name_combobox.set('')
        hf_models = ConfigHandler.CONFIGURATION_VALUES['huggingface_models']
        py_models = ConfigHandler.CONFIGURATION_VALUES['pytorch_models']
        tf_models = ConfigHandler.CONFIGURATION_VALUES['tensorflow_models']
        model_list = []

        self.name_label.grid_remove()
        self.name_combobox.grid_remove()
        self.lbp_name_entry.grid_remove()
        self.resize_height_label.grid_remove()
        self.resize_height_entry.grid_remove()
        self.resize_width_label.grid_remove()
        self.resize_width_entry.grid_remove()
        self.lbp_radius_label.grid_remove()
        self.lbp_radius_entry.grid_remove()
        self.lbp_num_points_label.grid_remove()
        self.lbp_num_points_entry.grid_remove()
        self.lbp_method_label.grid_remove()
        self.lbp_method_combobox.grid_remove()
        self.name_combobox.grid_remove()
        self.other_name_entry.grid_remove()
        self.other_name_label.grid_remove()
        self.custom_model_info_label.grid_remove()

        if framework == Framework.OPENCV.value:
            self.name_label.grid(row=0, column=1, padx=10, pady=5)
            self.lbp_name_entry.grid(row=1, column=1, padx=10, pady=5)
            self.resize_height_label.grid(row=0, column=2, padx=10)
            self.resize_height_entry.grid(row=1, column=2, padx=10, pady=5)
            self.resize_width_label.grid(row=0, column=3, padx=10)
            self.resize_width_entry.grid(row=1, column=3, padx=10, pady=5)
            self.lbp_radius_label.grid(row=2, column=1, padx=10)
            self.lbp_radius_entry.grid(row=3, column=1, padx=10, pady=5)
            self.lbp_num_points_label.grid(row=2, column=2, padx=10)
            self.lbp_num_points_entry.grid(row=3, column=2, padx=10, pady=5)
            self.lbp_method_label.grid(row=2, column=3, padx=10)
            self.lbp_method_combobox.grid(row=3, column=3, padx=10, pady=5)
            return

        elif framework == Framework.HUGGINGFACE.value:
            
            model_list = hf_models
            self.name_label.grid(row=0, column=1, padx=10, sticky="ew")
            self.name_combobox.grid(row=1, column=1, padx=10, pady=5, columnspan=2, sticky="ew")

        else:

            if framework == Framework.PYTORCH.value:
                model_list = py_models
            elif framework == Framework.TENSORFLOW.value:
                model_list = tf_models 

            self.name_label.grid(row=0, column=1, padx=10, sticky="ew")
            self.name_combobox.grid(row=1, column=1, padx=10, pady=5, columnspan=2, sticky="ew")
            self.resize_height_label.grid(row=0, column=3, padx=10)
            self.resize_height_entry.grid(row=1, column=3, padx=10, pady=5)
            self.resize_width_label.grid(row=0, column=4, padx=10)
            self.resize_width_entry.grid(row=1, column=4, padx=10, pady=5)

        self.name_combobox["values"] = model_list

        max_length = max([len(model) for model in model_list])  
        self.name_combobox.config(width=max_length + 5)
        return

    def validate_numeric(self, value):
         return value.isdigit() or value == ""
    
    def open_model_form(self, edit=False):

        added_models = self.clustering_data["clustering_models"].keys()

        if not edit:
            available_models = [model for model in ConfigHandler.CONFIGURATION_VALUES["clustering_models"] if model not in added_models]

            if not available_models:
                tk.messagebox.showinfo("Info", "All available models have already been added.")
                return
                
        else:
            available_models = [model for model in ConfigHandler.CONFIGURATION_VALUES["clustering_models"] if model in added_models]
            if not available_models:
                tk.messagebox.showinfo("Info", "No models have been added.")
                return


        model_form = ClusteringModelForm(self.top, allow_grid=True, models=available_models, config_data=self.clustering_data["clustering_models"], edit=edit)

        model_form.top.transient(self.top)
        model_form.top.grab_set()

        FormUtils.center_window(model_form, self.top)
        
        model_form.top.focus_set()

        self.top.wait_window(model_form.top)

        if model_form.config_data:
            self.add_model(model_form.config_data, "clustering_models")
            return
        
        self.clustering_models_text.config(state="normal")  
        self.clustering_models_text.delete(1.0, tk.END) 
        self.clustering_models_text.config(state="disabled")

    def add_model(self, model_params, param_key):

        model_name = list(model_params.keys())[0]
        
        self.clustering_data[param_key][model_name] = model_params[model_name]
            
        model_names = list(self.clustering_data[param_key].keys())
        models_text = "\n".join(model_names)
            
        self.clustering_models_text.config(state="normal")  
        self.clustering_models_text.delete(1.0, tk.END)  
        self.clustering_models_text.insert(tk.END, models_text)
        self.clustering_models_text.config(state="disabled")

        if model_names:
            self.edit_model_bt.grid()
        else:
            self.edit_model_bt.grid_remove()

    def check_embedding_model(self, model_name, framework):
        if framework.lower() == Framework.HUGGINGFACE.value:
            try:
                model_info = huggingface_hub.model_info(model_name)
                return True
            except Exception as e:
                return False
            
        if framework.lower() == Framework.PYTORCH.value:
            try:
                model = getattr(models, model_name)
                return True
            except AttributeError:
                return False
            
        if framework.lower() == Framework.TENSORFLOW.value:
            try:
                model = getattr(tf.keras.applications, model_name)
                return True
            except AttributeError:
                return False
        

    def validate_form(self):

        embedding_model_info = self.clustering_data["embedding_model"]
        framework = embedding_model_info["framework"].get()

        if framework == Framework.OPENCV.value:

            if embedding_model_info["resize_height"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", "Resize height must be specified when using an opencv feature extractor model.")
                return False
            
            if embedding_model_info["resize_width"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", "Resize width must be specified when using an opencv feature extractor model.")
                return False
            
            if embedding_model_info["lbp_radius"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", "LBP radius must be specified when using an opencv feature extractor model.")
                return False
            
            if embedding_model_info["lbp_num_points"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", "Num points for LBP must be specified when using an opencv feature extractor model.")
                return False
            
        elif framework == Framework.HUGGINGFACE.value:

            if embedding_model_info["name"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", f"The name of the {framework} embedding model to use must be specified.")
                return False

        elif framework == Framework.PYTORCH.value:

            if embedding_model_info["name"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", f"The name of the {framework} embedding model to use must be specified.")
                return False
            
            if embedding_model_info["resize_height"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", f"Resize height must be specified when using a {framework} embedding model.")
                return False
            
            if embedding_model_info["resize_width"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", f"Resize width must be specified when using a {framework} embedding model.")
                return False
            
        elif framework == Framework.TENSORFLOW.value:

            if embedding_model_info["name"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", f"The name of the {framework} embedding model to use must be specified.")
                return False
            
            if embedding_model_info["resize_height"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", f"Resize height must be specified when using a {framework} embedding model.")
                return False
            
            if embedding_model_info["resize_width"].get().strip() == "":
                tk.messagebox.showerror("Clustering configuration error", f"Resize width must be specified when using a {framework} embedding model.")
                return False
            
        if embedding_model_info["name"].get().lower() == "other":

            custom_model_name = embedding_model_info["name_other"].get()
                
            if custom_model_name.strip() == "":
                tk.messagebox.showerror("Clustering configuration error", "Custom model name must be specified.")
                return False
            
            if not self.check_embedding_model(custom_model_name, framework):
                tk.messagebox.showerror("Clustering configuration error", f"Embedding model name {custom_model_name} does not exist for {framework} framework.")
                return False
         
            
        if not self.clustering_data["clustering_models"]:
            tk.messagebox.showerror("Clustering configuration error", "At least one clustering model must be defined.")
            return False
        
        if embedding_model_info["name"].get().lower() == "other":
            tk.messagebox.showwarning("Unverified model warning", f"The embedding model you selected from {framework} has not been verified with this application. There is a possibility of unexpected behavior or errors.\n\n"
                                      "If you experience issues, try selecting a model from the available list.")


        return True

    def next(self):
        if not self.validate_form():
            return
        
        self.config_data["clustering_data"].update(self.clustering_data)
        self.controller.show_frame("ReductionFrame")
